"""
Iridas .cube LUT Format Input / Output Utilities
================================================

Defines the *Iridas* *.cube* *LUT* format related input / output utilities
objects:

-   :func:`colour.io.read_LUT_IridasCube`
-   :func:`colour.io.write_LUT_IridasCube`

References
----------
-   :cite:`AdobeSystems2013b` : Adobe Systems. (2013). Cube LUT Specification.
    https://drive.google.com/open?id=143Eh08ZYncCAMwJ1q4gWxVOqR_OSWYvs
"""

from __future__ import annotations

import numpy as np

from colour.io.luts import LUT1D, LUT3x1D, LUT3D, LUTSequence
from colour.io.luts.common import path_to_title
from colour.hints import Boolean, Integer, List, Tuple, Union
from colour.utilities import (
    as_float_array,
    as_int_scalar,
    attest,
    usage_warning,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "read_LUT_IridasCube",
    "write_LUT_IridasCube",
]


def read_LUT_IridasCube(path: str) -> Union[LUT3x1D, LUT3D]:
    """
    Read given *Iridas* *.cube* *LUT* file.

    Parameters
    ----------
    path
        *LUT* path.

    Returns
    -------
    :class:`LUT3x1D` or :class:`LUT3D`.
        :class:`LUT3x1D` or :class:`LUT3D` class instance.

    References
    ----------
    :cite:`AdobeSystems2013b`

    Examples
    --------
    Reading a 3x1D *Iridas* *.cube* *LUT*:

    >>> import os
    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'iridas_cube',
    ...     'ACES_Proxy_10_to_ACES.cube')
    >>> print(read_LUT_IridasCube(path))
    LUT3x1D - ACES Proxy 10 to ACES
    -------------------------------
    <BLANKLINE>
    Dimensions : 2
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (32, 3)

    Reading a 3D *Iridas* *.cube* *LUT*:

    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'iridas_cube',
    ...     'Colour_Correct.cube')
    >>> print(read_LUT_IridasCube(path))
    LUT3D - Generated by Foundry::LUT
    ---------------------------------
    <BLANKLINE>
    Dimensions : 3
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (4, 4, 4, 3)

    Reading a 3D *Iridas* *.cube* *LUT* with comments:

    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'iridas_cube',
    ...     'Demo.cube')
    >>> print(read_LUT_IridasCube(path))
    LUT3x1D - Demo
    --------------
    <BLANKLINE>
    Dimensions : 2
    Domain     : [[ 0.  0.  0.]
                  [ 1.  2.  3.]]
    Size       : (3, 3)
    Comment 01 : Comments can go anywhere
    """

    title = path_to_title(path)
    domain_min, domain_max = np.array([0, 0, 0]), np.array([1, 1, 1])
    dimensions: Integer = 3
    size: Integer = 2
    data = []
    comments = []

    with open(path) as cube_file:
        lines = cube_file.readlines()
        for line in lines:
            line = line.strip()

            if len(line) == 0:
                continue

            if line.startswith("#"):
                comments.append(line[1:].strip())
                continue

            tokens = line.split()
            if tokens[0] == "TITLE":
                title = " ".join(tokens[1:])[1:-1]
            elif tokens[0] == "DOMAIN_MIN":
                domain_min = as_float_array(tokens[1:])
            elif tokens[0] == "DOMAIN_MAX":
                domain_max = as_float_array(tokens[1:])
            elif tokens[0] == "LUT_1D_SIZE":
                dimensions = 2
                size = as_int_scalar(tokens[1])
            elif tokens[0] == "LUT_3D_SIZE":
                dimensions = 3
                size = as_int_scalar(tokens[1])
            else:
                data.append(tokens)

    table = as_float_array(data)

    LUT: Union[LUT3x1D, LUT3D]
    if dimensions == 2:
        LUT = LUT3x1D(
            table,
            title,
            np.vstack([domain_min, domain_max]),
            comments=comments,
        )
    elif dimensions == 3:
        # The lines of table data shall be in ascending index order,
        # with the first component index (Red) changing most rapidly,
        # and the last component index (Blue) changing least rapidly.
        table = table.reshape([size, size, size, 3], order="F")

        LUT = LUT3D(
            table,
            title,
            np.vstack([domain_min, domain_max]),
            comments=comments,
        )

    return LUT


def write_LUT_IridasCube(
    LUT: Union[LUT3x1D, LUT3D, LUTSequence], path: str, decimals: Integer = 7
) -> Boolean:
    """
    Write given *LUT* to given  *Iridas* *.cube* *LUT* file.

    Parameters
    ----------
    LUT
        :class:`LUT3x1D`, :class:`LUT3D` or :class:`LUTSequence` class instance
        to write at given path.
    path
        *LUT* path.
    decimals
        Formatting decimals.

    Returns
    -------
    :class:`bool`
        Definition success.

    Warnings
    --------
    -   If a :class:`LUTSequence` class instance is passed as ``LUT``, the
        first *LUT* in the *LUT* sequence will be used.

    References
    ----------
    :cite:`AdobeSystems2013b`

    Examples
    --------
    Writing a 3x1D *Iridas* *.cube* *LUT*:

    >>> from colour.algebra import spow
    >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
    >>> LUT = LUT3x1D(
    ...     spow(LUT3x1D.linear_table(16, domain), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.'])
    >>> write_LUT_IridasCube(LUTxD, 'My_LUT.cube')  # doctest: +SKIP

    Writing a 3D *Iridas* *.cube* *LUT*:

    >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
    >>> LUT = LUT3D(
    ...     spow(LUT3D.linear_table(16, domain), 1 / 2.2),
    ...     'My LUT',
    ...     np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]]),
    ...     comments=['A first comment.', 'A second comment.'])
    >>> write_LUT_IridasCube(LUTxD, 'My_LUT.cube')  # doctest: +SKIP
    """

    if isinstance(LUT, LUTSequence):
        usage_warning(
            f'"LUT" is a "LUTSequence" instance was passed, '
            f'using first sequence "LUT":\n{LUT}'
        )
        LUTxD = LUT[0]
    elif isinstance(LUT, LUT1D):
        LUTxD = LUT.as_LUT(LUT3x1D)
    else:
        LUTxD = LUT

    attest(
        isinstance(LUTxD, (LUT3x1D, LUT3D)),
        '"LUT" must be a 1D, 3x1D or 3D "LUT"!',
    )

    attest(not LUTxD.is_domain_explicit(), '"LUT" domain must be implicit!')

    is_3x1D = isinstance(LUTxD, LUT3x1D)

    size = LUTxD.size
    if is_3x1D:
        attest(2 <= size <= 65536, '"LUT" size must be in domain [2, 65536]!')
    else:
        attest(2 <= size <= 256, '"LUT" size must be in domain [2, 256]!')

    def _format_array(array: Union[List, Tuple]) -> str:
        """Format given array as an *Iridas* *.cube* data row."""

        return "{1:0.{0}f} {2:0.{0}f} {3:0.{0}f}".format(decimals, *array)

    with open(path, "w") as cube_file:
        cube_file.write(f'TITLE "{LUTxD.name}"\n')

        if LUTxD.comments:
            for comment in LUTxD.comments:
                cube_file.write(f"# {comment}\n")

        cube_file.write(
            f"{'LUT_1D_SIZE' if is_3x1D else 'LUT_3D_SIZE'} {LUTxD.table.shape[0]}\n"
        )

        default_domain = np.array([[0, 0, 0], [1, 1, 1]])
        if not np.array_equal(LUTxD.domain, default_domain):
            cube_file.write(f"DOMAIN_MIN {_format_array(LUTxD.domain[0])}\n")
            cube_file.write(f"DOMAIN_MAX {_format_array(LUTxD.domain[1])}\n")

        if not is_3x1D:
            table = LUTxD.table.reshape([-1, 3], order="F")
        else:
            table = LUTxD.table

        for row in table:
            cube_file.write(f"{_format_array(row)}\n")

    return True



"""
Pal/Secam Colourspace
=====================

Defines the *Pal/Secam* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_PAL_SECAM`.

References
----------
-   :cite:`InternationalTelecommunicationUnion1998a` : International
    Telecommunication Union. (1998). Recommendation ITU-R BT.470-6 -
    CONVENTIONAL TELEVISION SYSTEMS (pp. 1-36).
    http://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.470-6-199811-S!!PDF-E.pdf
"""

from __future__ import annotations

from colour.hints import NDArray
from colour.models.rgb import RGB_Colourspace
from colour.models.rgb.datasets.itur_bt_470 import (
    PRIMARIES_BT470_625,
    CCS_WHITEPOINT_BT470_625,
    WHITEPOINT_NAME_BT470_625,
    MATRIX_BT470_625_TO_XYZ,
    MATRIX_XYZ_TO_BT470_625,
    RGB_COLOURSPACE_BT470_625,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_PAL_SECAM",
    "WHITEPOINT_NAME_PAL_SECAM",
    "CCS_WHITEPOINT_PAL_SECAM",
    "MATRIX_PAL_SECAM_TO_XYZ",
    "MATRIX_XYZ_TO_PAL_SECAM",
    "RGB_COLOURSPACE_PAL_SECAM",
]

PRIMARIES_PAL_SECAM: NDArray = PRIMARIES_BT470_625
"""*Pal/Secam* colourspace primaries."""

WHITEPOINT_NAME_PAL_SECAM: str = WHITEPOINT_NAME_BT470_625
"""*Pal/Secam* colourspace whitepoint name."""

CCS_WHITEPOINT_PAL_SECAM: NDArray = CCS_WHITEPOINT_BT470_625
"""*Pal/Secam* colourspace whitepoint chromaticity coordinates."""

MATRIX_PAL_SECAM_TO_XYZ: NDArray = MATRIX_BT470_625_TO_XYZ
"""*Pal/Secam* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_PAL_SECAM: NDArray = MATRIX_XYZ_TO_BT470_625
"""*CIE XYZ* tristimulus values to *Pal/Secam* colourspace matrix."""

RGB_COLOURSPACE_PAL_SECAM: RGB_Colourspace = RGB_Colourspace(
    "Pal/Secam",
    PRIMARIES_PAL_SECAM,
    CCS_WHITEPOINT_PAL_SECAM,
    WHITEPOINT_NAME_PAL_SECAM,
    MATRIX_PAL_SECAM_TO_XYZ,
    MATRIX_XYZ_TO_PAL_SECAM,
    RGB_COLOURSPACE_BT470_625.cctf_encoding,
    RGB_COLOURSPACE_BT470_625.cctf_decoding,
)
RGB_COLOURSPACE_PAL_SECAM.__doc__ = """
*Pal/Secam* colourspace.

References
----------
:cite:`InternationalTelecommunicationUnion1998a`
"""



"""
Y'CbCr Colour Encoding
======================

Defines the *Y'CbCr* colour encoding related attributes and objects:

-   :attr:`colour.WEIGHTS_YCBCR`
-   :func:`colour.matrix_YCbCr`
-   :func:`colour.offset_YCbCr`
-   :func:`colour.RGB_to_YCbCr`
-   :func:`colour.YCbCr_to_RGB`
-   :func:`colour.RGB_to_YcCbcCrc`
-   :func:`colour.YcCbcCrc_to_RGB`

Notes
-----
-   *Y'CbCr* is not an absolute colourspace.

References
----------
-   :cite:`InternationalTelecommunicationUnion2011e` : International
    Telecommunication Union. (2011). Recommendation ITU-T T.871 - Information
    technology - Digital compression and coding of continuous-tone still
    images: JPEG File Interchange Format (JFIF).
    https://www.itu.int/rec/dologin_pub.asp?lang=e&\
id=T-REC-T.871-201105-I!!PDF-E&type=items
-   :cite:`InternationalTelecommunicationUnion2015h` : International
    Telecommunication Union. (2015). Recommendation ITU-R BT.2020 - Parameter
    values for ultra-high definition television systems for production and
    international programme exchange (pp. 1-8).
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.2020-2-201510-I!!PDF-E.pdf
-   :cite:`InternationalTelecommunicationUnion2015i` : International
    Telecommunication Union. (2015). Recommendation ITU-R BT.709-6 - Parameter
    values for the HDTV standards for production and international programme
    exchange BT Series Broadcasting service (pp. 1-32).
    https://www.itu.int/dms_pubrec/itu-r/rec/bt/\
R-REC-BT.709-6-201506-I!!PDF-E.pdf
-   :cite:`SocietyofMotionPictureandTelevisionEngineers1999b` : Society of
    Motion Picture and Television Engineers. (1999). ANSI/SMPTE 240M-1995 -
    Signal Parameters - 1125-Line High-Definition Production Systems (pp. 1-7).
    http://car.france3.mars.free.fr/HD/\
INA-%2026%20jan%2006/SMPTE%20normes%20et%20confs/s240m.pdf
-   :cite:`Wikipedia2004d` : Wikipedia. (2004). YCbCr. Retrieved February 29,
    2016, from https://en.wikipedia.org/wiki/YCbCr
"""

from __future__ import annotations

import numpy as np

from colour.hints import Any, ArrayLike, Boolean, Integer, NDArray
from colour.models.rgb.transfer_functions import (
    CV_range,
    eotf_inverse_BT2020,
    eotf_BT2020,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_float_array,
    as_int_array,
    domain_range_scale,
    from_range_1,
    to_domain_1,
    tsplit,
    tstack,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Development"

__all__ = [
    "WEIGHTS_YCBCR",
    "ranges_YCbCr",
    "matrix_YCbCr",
    "offset_YCbCr",
    "RGB_to_YCbCr",
    "YCbCr_to_RGB",
    "RGB_to_YcCbcCrc",
    "YcCbcCrc_to_RGB",
]

WEIGHTS_YCBCR: CaseInsensitiveMapping = CaseInsensitiveMapping(
    {
        "ITU-R BT.601": np.array([0.2990, 0.1140]),
        "ITU-R BT.709": np.array([0.2126, 0.0722]),
        "ITU-R BT.2020": np.array([0.2627, 0.0593]),
        "SMPTE-240M": np.array([0.2122, 0.0865]),
    }
)
"""
Luma weightings presets.

References
----------
:cite:`InternationalTelecommunicationUnion2011e`,
:cite:`InternationalTelecommunicationUnion2015i`,
:cite:`InternationalTelecommunicationUnion2015h`,
:cite:`SocietyofMotionPictureandTelevisionEngineers1999b`,
:cite:`Wikipedia2004d`
"""


def ranges_YCbCr(bits: Integer, is_legal: Boolean, is_int: Boolean) -> NDArray:
    """
    Return the *Y'CbCr* colour encoding ranges array for given bit depth,
    range legality and representation.

    Parameters
    ----------
    bits
        Bit depth of the *Y'CbCr* colour encoding ranges array.
    is_legal
        Whether the *Y'CbCr* colour encoding ranges array is legal.
    is_int
        Whether the *Y'CbCr* colour encoding ranges array represents integer
        code values.

    Returns
    -------
    :class:`numpy.ndarray`
        *Y'CbCr* colour encoding ranges array.

    Examples
    --------
    >>> ranges_YCbCr(8, True, True)
    array([ 16, 235,  16, 240])
    >>> ranges_YCbCr(8, True, False)  # doctest: +ELLIPSIS
    array([ 0.0627451...,  0.9215686...,  0.0627451...,  0.9411764...])
    >>> ranges_YCbCr(10, False, False)
    array([ 0. ,  1. , -0.5,  0.5])
    """

    if is_legal:
        ranges = np.array([16, 235, 16, 240])
        ranges *= 2 ** (bits - 8)
    else:
        ranges = np.array([0, 2**bits - 1, 0, 2**bits - 1])

    if not is_int:
        ranges = as_int_array(ranges) / (2**bits - 1)

    if is_int and not is_legal:
        ranges[3] = 2**bits

    if not is_int and not is_legal:
        ranges[2] = -0.5
        ranges[3] = 0.5

    return ranges


def matrix_YCbCr(
    K: NDArray = WEIGHTS_YCBCR["ITU-R BT.709"],
    bits: Integer = 8,
    is_legal: Boolean = False,
    is_int: Boolean = False,
) -> NDArray:
    """
    Compute the *R'G'B'* to *Y'CbCr* matrix for given weights, bit depth,
    range legality and representation.

    The related offset for the *R'G'B'* to *Y'CbCr* matrix can be computed with
    the :func:`colour.offset_YCbCr` definition.

    Parameters
    ----------
    K
        Luma weighting coefficients of red and blue. See
        :attr:`colour.WEIGHTS_YCBCR` for presets. Default is
        *(0.2126, 0.0722)*, the weightings for *ITU-R BT.709*.
    bits
        Bit depth of the *Y'CbCr* colour encoding ranges array.
    is_legal
        Whether the *Y'CbCr* colour encoding ranges array is legal.
    is_int
        Whether the *Y'CbCr* colour encoding ranges array represents integer
        code values.

    Returns
    -------
    :class:`numpy.ndarray`
        *Y'CbCr* matrix.

    Examples
    --------
    >>> matrix_YCbCr()  # doctest: +ELLIPSIS
    array([[  1.0000000...e+00,  ...,   1.5748000...e+00],
           [  1.0000000...e+00,  -1.8732427...e-01,  -4.6812427...e-01],
           [  1.0000000...e+00,   1.8556000...e+00,  ...]])
    >>> matrix_YCbCr(K=WEIGHTS_YCBCR['ITU-R BT.601'])  # doctest: +ELLIPSIS
    array([[  1.0000000...e+00,  ...,   1.4020000...e+00],
           [  1.0000000...e+00,  -3.4413628...e-01,  -7.1413628...e-01],
           [  1.0000000...e+00,   1.7720000...e+00,  ...]])
    >>> matrix_YCbCr(is_legal=True)  # doctest: +ELLIPSIS
    array([[  1.1643835...e+00,  ...,   1.7927410...e+00],
           [  1.1643835...e+00,  -2.1324861...e-01,  -5.3290932...e-01],
           [  1.1643835...e+00,   2.1124017...e+00,  ...]])

    Matching the default output of the :func:`colour.RGB_to_YCbCr` is done as
    follows:

    >>> from colour.algebra import vector_dot
    >>> from colour.utilities import as_int_array
    >>> RGB = np.array([1.0, 1.0, 1.0])
    >>> RGB_to_YCbCr(RGB)  # doctest: +ELLIPSIS
    array([ 0.9215686...,  0.5019607...,  0.5019607...])
    >>> YCbCr = vector_dot(np.linalg.inv(matrix_YCbCr(is_legal=True)), RGB)
    >>> YCbCr += offset_YCbCr(is_legal=True)
    >>> YCbCr  # doctest: +ELLIPSIS
    array([ 0.9215686...,  0.5019607...,  0.5019607...])

    Matching the int output of the :func:`colour.RGB_to_YCbCr` is done as
    follows:

    >>> RGB = np.array([102, 0, 51])
    >>> RGB_to_YCbCr(RGB, in_bits=8, in_int=True, out_bits=8, out_int=True)
    ... # doctest: +SKIP
    array([ 38, 140, 171])
    >>> YCbCr = vector_dot(np.linalg.inv(matrix_YCbCr(is_legal=True)), RGB)
    >>> YCbCr += offset_YCbCr(is_legal=True, is_int=True)
    >>> as_int_array(np.around(YCbCr))
    ... # doctest: +SKIP
    array([ 38, 140, 171])
    """

    Kr, Kb = K
    Y_min, Y_max, C_min, C_max = ranges_YCbCr(bits, is_legal, is_int)

    Y = np.array([Kr, (1 - Kr - Kb), Kb])
    Cb = 0.5 * (np.array([0, 0, 1]) - Y) / (1 - Kb)
    Cr = 0.5 * (np.array([1, 0, 0]) - Y) / (1 - Kr)
    Y *= Y_max - Y_min
    Cb *= C_max - C_min
    Cr *= C_max - C_min

    return np.linalg.inv(np.vstack([Y, Cb, Cr]))


def offset_YCbCr(
    bits: Integer = 8, is_legal: Boolean = False, is_int: Boolean = False
) -> NDArray:
    """
    Compute the *R'G'B'* to *Y'CbCr* offsets for given bit depth, range
    legality and representation.

    The related *R'G'B'* to *Y'CbCr* matrix can be computed with the
    :func:`colour.matrix_YCbCr` definition.

    Parameters
    ----------
    bits
        Bit depth of the *Y'CbCr* colour encoding ranges array.
    is_legal
        Whether the *Y'CbCr* colour encoding ranges array is legal.
    is_int
        Whether the *Y'CbCr* colour encoding ranges array represents integer
        code values.

    Returns
    -------
    :class:`numpy.ndarray`
        *Y'CbCr* matrix.

    Examples
    --------
    >>> offset_YCbCr()
    array([ 0.,  0.,  0.])
    >>> offset_YCbCr(is_legal=True)  # doctest: +ELLIPSIS
    array([ 0.0627451...,  0.5019607...,  0.5019607...])
    """

    Y_min, _Y_max, C_min, C_max = ranges_YCbCr(bits, is_legal, is_int)

    Y_offset = Y_min
    C_offset = (C_min + C_max) / 2

    return np.array([Y_offset, C_offset, C_offset])


def RGB_to_YCbCr(
    RGB: ArrayLike,
    K: NDArray = WEIGHTS_YCBCR["ITU-R BT.709"],
    in_bits: Integer = 10,
    in_legal: Boolean = False,
    in_int: Boolean = False,
    out_bits: Integer = 8,
    out_legal: Boolean = True,
    out_int: Boolean = False,
    **kwargs: Any,
) -> NDArray:
    """
    Convert an array of *R'G'B'* values to the corresponding *Y'CbCr* colour
    encoding values array.

    Parameters
    ----------
    RGB
        Input *R'G'B'* array of floats or integer values.
    K
        Luma weighting coefficients of red and blue. See
        :attr:`colour.WEIGHTS_YCBCR` for presets. Default is
        *(0.2126, 0.0722)*, the weightings for *ITU-R BT.709*.
    in_bits
        Bit depth for integer input, or used in the calculation of the
        denominator for legal range float values, i.e. 8-bit means the float
        value for legal white is *235 / 255*. Default is *10*.
    in_legal
        Whether to treat the input values as legal range. Default is *False*.
    in_int
        Whether to treat the input values as ``in_bits`` integer code values.
        Default is *False*.
    out_bits
        Bit depth for integer output, or used in the calculation of the
        denominator for legal range float values, i.e. 8-bit means the float
        value for legal white is *235 / 255*. Ignored if ``out_legal`` and
        ``out_int`` are both *False*. Default is *8*.
    out_legal
        Whether to return legal range values. Default is *True*.
    out_int
        Whether to return values as ``out_bits`` integer code values. Default
        is *False*.

    Other Parameters
    ----------------
    in_range
        Array overriding the computed range such as
        *in_range = (RGB_min, RGB_max)*. If ``in_range`` is undefined,
        *RGB_min* and *RGB_max* will be computed using :func:`colour.CV_range`
        definition.
    out_range
        Array overriding the computed range such as
        *out_range = (Y_min, Y_max, C_min, C_max)`. If ``out_range`` is
        undefined, *Y_min*, *Y_max*, *C_min* and *C_max* will be computed
        using :func:`colour.models.rgb.ycbcr.ranges_YCbCr` definition.

    Returns
    -------
    :class:`numpy.ndarray`
        *Y'CbCr* colour encoding array of integer or float values.

    Warnings
    --------
    For *Recommendation ITU-R BT.2020*, :func:`colour.RGB_to_YCbCr` definition
    is only applicable to the non-constant luminance implementation.
    :func:`colour.RGB_to_YcCbcCrc` definition should be used for the constant
    luminance case as per :cite:`InternationalTelecommunicationUnion2015h`.

    Notes
    -----
    +----------------+-----------------------+---------------+
    | **Domain \\***  | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``RGB``        | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    +----------------+-----------------------+---------------+
    | **Range \\***   | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``YCbCr``      | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    \\* This definition has input and output integer switches, thus the
    domain-range scale information is only given for the floating point mode.

    -   The default arguments, ``**{'in_bits': 10, 'in_legal': False,
        'in_int': False, 'out_bits': 8, 'out_legal': True, 'out_int': False}``
        transform a float *R'G'B'* input array normalised to domain [0, 1]
        (``in_bits`` is ignored) to a float *Y'CbCr* output array where *Y'* is
        normalised to range [16 / 255, 235 / 255] and *Cb* and *Cr* are
        normalised to range [16 / 255, 240./255]. The float values are
        calculated based on an [0, 255] integer range, but no 8-bit
        quantisation or clamping are performed.

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2011e`,
    :cite:`InternationalTelecommunicationUnion2015i`,
    :cite:`SocietyofMotionPictureandTelevisionEngineers1999b`,
    :cite:`Wikipedia2004d`

    Examples
    --------
    >>> RGB = np.array([1.0, 1.0, 1.0])
    >>> RGB_to_YCbCr(RGB)  # doctest: +ELLIPSIS
    array([ 0.9215686...,  0.5019607...,  0.5019607...])

    Matching the float output of *The Foundry Nuke*'s *Colorspace* node set to
    *YCbCr*:

    >>> RGB_to_YCbCr(RGB,
    ...              out_range=(16 / 255, 235 / 255, 15.5 / 255, 239.5 / 255))
    ... # doctest: +ELLIPSIS
    array([ 0.9215686...,  0.5       ,  0.5       ])

    Matching the float output of *The Foundry Nuke*'s *Colorspace* node set to
    *YPbPr*:

    >>> RGB_to_YCbCr(RGB, out_legal=False, out_int=False)
    ... # doctest: +ELLIPSIS
    array([ 1.,  0.,  0.])

    Creating integer code values as per standard *10-bit SDI*:

    >>> RGB_to_YCbCr(RGB, out_legal=True, out_bits=10, out_int=True)
    ... # doctest: +ELLIPSIS
    array([940, 512, 512]...)

    For *JFIF JPEG* conversion as per *Recommendation ITU-T T.871*

    >>> RGB = np.array([102, 0, 51])
    >>> RGB_to_YCbCr(RGB, K=WEIGHTS_YCBCR['ITU-R BT.601'], in_range=(0, 255),
    ...              out_range=(0, 255, 0, 256), out_int=True)
    ... # doctest: +ELLIPSIS
    array([ 36, 136, 175]...)

    Note the use of 256 for the max *Cb / Cr* value, which is required so that
    the *Cb* and *Cr* output is centered about 128. Using 255 centres it
    about 127.5, meaning that there is no integer code value to represent
    achromatic colours. This does however create the possibility of output
    integer codes with value of 256, which cannot be stored in 8-bit integer
    representation. *Recommendation ITU-T T.871* specifies these should be
    clamped to 255.

    These *JFIF JPEG* ranges are also obtained as follows:

    >>> RGB_to_YCbCr(RGB, K=WEIGHTS_YCBCR['ITU-R BT.601'], in_bits=8,
    ...              in_int=True, out_legal=False, out_int=True)
    ... # doctest: +ELLIPSIS
    array([ 36, 136, 175]...)
    """

    if in_int:
        RGB = as_float_array(RGB)
    else:
        RGB = to_domain_1(RGB)

    Kr, Kb = K
    RGB_min, RGB_max = kwargs.get(
        "in_range", CV_range(in_bits, in_legal, in_int)
    )
    Y_min, Y_max, C_min, C_max = kwargs.get(
        "out_range", ranges_YCbCr(out_bits, out_legal, out_int)
    )

    RGB_float = as_float_array(RGB) - RGB_min
    RGB_float *= 1 / (RGB_max - RGB_min)
    R, G, B = tsplit(RGB_float)

    Y = Kr * R + (1 - Kr - Kb) * G + Kb * B
    Cb = 0.5 * (B - Y) / (1 - Kb)
    Cr = 0.5 * (R - Y) / (1 - Kr)
    Y *= Y_max - Y_min
    Y += Y_min
    Cb *= C_max - C_min
    Cr *= C_max - C_min
    Cb += (C_max + C_min) / 2
    Cr += (C_max + C_min) / 2

    YCbCr = tstack([Y, Cb, Cr])

    if out_int:
        return as_int_array(np.round(YCbCr))
    else:
        return from_range_1(YCbCr)


def YCbCr_to_RGB(
    YCbCr: ArrayLike,
    K: NDArray = WEIGHTS_YCBCR["ITU-R BT.709"],
    in_bits: Integer = 8,
    in_legal: Boolean = True,
    in_int: Boolean = False,
    out_bits: Integer = 10,
    out_legal: Boolean = False,
    out_int: Boolean = False,
    **kwargs: Any,
) -> NDArray:
    """
    Convert an array of *Y'CbCr* colour encoding values to the corresponding
    *R'G'B'* values array.

    Parameters
    ----------
    YCbCr
        Input *Y'CbCr* colour encoding array of integer or float values.
    K
        Luma weighting coefficients of red and blue. See
        :attr:`colour.WEIGHTS_YCBCR` for presets. Default is
        *(0.2126, 0.0722)*, the weightings for *ITU-R BT.709*.
    in_bits
        Bit depth for integer input, or used in the calculation of the
        denominator for legal range float values, i.e. 8-bit means the float
        value for legal white is *235 / 255*. Default is *8*.
    in_legal
        Whether to treat the input values as legal range. Default is *True*.
    in_int
        Whether to treat the input values as ``in_bits`` integer code values.
        Default is *False*.
    out_bits
        Bit depth for integer output, or used in the calculation of the
        denominator for legal range float values, i.e. 8-bit means the float
        value for legal white is *235 / 255*. Ignored if ``out_legal`` and
        ``out_int`` are both *False*. Default is *10*.
    out_legal
        Whether to return legal range values. Default is *False*.
    out_int
        Whether to return values as ``out_bits`` integer code values. Default
        is *False*.

    Other Parameters
    ----------------
    in_range
        Array overriding the computed range such as
        *in_range = (Y_min, Y_max, C_min, C_max)*. If ``in_range`` is
        undefined, *Y_min*, *Y_max*, *C_min* and *C_max* will be computed using
        :func:`colour.models.rgb.ycbcr.ranges_YCbCr` definition.
    out_range
        Array overriding the computed range such as
        *out_range = (RGB_min, RGB_max)*. If ``out_range`` is undefined,
        *RGB_min* and *RGB_max* will be computed using :func:`colour.CV_range`
        definition.

    Returns
    -------
    :class:`numpy.ndarray`
        *R'G'B'* array of integer or float values.

    Notes
    -----
    +----------------+-----------------------+---------------+
    | **Domain \\***  | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``YCbCr``      | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    +----------------+-----------------------+---------------+
    | **Range \\***   | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``RGB``        | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    \\* This definition has input and output integer switches, thus the
    domain-range scale information is only given for the floating point mode.

    Warnings
    --------
    For *Recommendation ITU-R BT.2020*, :func:`colour.YCbCr_to_RGB`
    definition is only applicable to the non-constant luminance implementation.
    :func:`colour.YcCbcCrc_to_RGB` definition should be used for the constant
    luminance case as per :cite:`InternationalTelecommunicationUnion2015h`.

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2011e`,
    :cite:`InternationalTelecommunicationUnion2015i`,
    :cite:`SocietyofMotionPictureandTelevisionEngineers1999b`,
    :cite:`Wikipedia2004d`

    Examples
    --------
    >>> YCbCr = np.array([502, 512, 512])
    >>> YCbCr_to_RGB(YCbCr, in_bits=10, in_legal=True, in_int=True)
    array([ 0.5,  0.5,  0.5])
    """

    if in_int:
        YCbCr = as_float_array(YCbCr)
    else:
        YCbCr = to_domain_1(YCbCr)

    Y, Cb, Cr = tsplit(YCbCr)
    Kr, Kb = K
    Y_min, Y_max, C_min, C_max = kwargs.get(
        "in_range", ranges_YCbCr(in_bits, in_legal, in_int)
    )
    RGB_min, RGB_max = kwargs.get(
        "out_range", CV_range(out_bits, out_legal, out_int)
    )

    Y -= Y_min
    Cb -= (C_max + C_min) / 2
    Cr -= (C_max + C_min) / 2
    Y *= 1 / (Y_max - Y_min)
    Cb *= 1 / (C_max - C_min)
    Cr *= 1 / (C_max - C_min)
    R = Y + (2 - 2 * Kr) * Cr
    B = Y + (2 - 2 * Kb) * Cb
    G = (Y - Kr * R - Kb * B) / (1 - Kr - Kb)

    RGB = tstack([R, G, B])
    RGB *= RGB_max - RGB_min
    RGB += RGB_min

    RGB = as_int_array(np.round(RGB)) if out_int else from_range_1(RGB)

    return RGB


def RGB_to_YcCbcCrc(
    RGB: ArrayLike,
    out_bits: Integer = 10,
    out_legal: Boolean = True,
    out_int: Boolean = False,
    is_12_bits_system: Boolean = False,
    **kwargs: Any,
) -> NDArray:
    """
    Convert an array of *RGB* linear values to the corresponding *Yc'Cbc'Crc'*
    colour encoding values array.

    Parameters
    ----------
    RGB
        Input *RGB* array of linear float values.
    out_bits
        Bit depth for integer output, or used in the calculation of the
        denominator for legal range float values, i.e. 8-bit means the float
        value for legal white is *235 / 255*. Ignored if ``out_legal`` and
        ``out_int`` are both *False*. Default is *10*.
    out_legal
        Whether to return legal range values. Default is *True*.
    out_int
        Whether to return values as ``out_bits`` integer code values. Default
        is *False*.
    is_12_bits_system
        *Recommendation ITU-R BT.2020* OETF (OECF) adopts different parameters
        for 10 and 12 bit systems. Default is *False*.

    Other Parameters
    ----------------
    out_range
        Array overriding the computed range such as
        *out_range = (Y_min, Y_max, C_min, C_max)*. If ``out_range`` is
        undefined, *Y_min*, *Y_max*, *C_min* and *C_max* will be computed
        using :func:`colour.models.rgb.ycbcr.ranges_YCbCr` definition.

    Returns
    -------
    :class:`numpy.ndarray`
        *Yc'Cbc'Crc'* colour encoding array of integer or float values.

    Notes
    -----
    +----------------+-----------------------+---------------+
    | **Domain \\***  | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``RGB``        | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    +----------------+-----------------------+---------------+
    | **Range \\***   | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``YcCbcCrc``   | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    \\* This definition has input and output integer switches, thus the
    domain-range scale information is only given for the floating point mode.

    Warnings
    --------
    This definition is specifically for usage with
    *Recommendation ITU-R BT.2020* when adopting the constant luminance
    implementation.

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2015h`, :cite:`Wikipedia2004d`

    Examples
    --------
    >>> RGB = np.array([0.18, 0.18, 0.18])
    >>> RGB_to_YcCbcCrc(RGB, out_legal=True, out_bits=10, out_int=True,
    ...                 is_12_bits_system=False)
    ... # doctest: +ELLIPSIS
    array([422, 512, 512]...)
    """

    R, G, B = tsplit(to_domain_1(RGB))
    Y_min, Y_max, C_min, C_max = kwargs.get(
        "out_range", ranges_YCbCr(out_bits, out_legal, out_int)
    )

    Yc = 0.2627 * R + 0.6780 * G + 0.0593 * B

    with domain_range_scale("ignore"):
        Yc = eotf_inverse_BT2020(Yc, is_12_bits_system=is_12_bits_system)
        R = eotf_inverse_BT2020(R, is_12_bits_system=is_12_bits_system)
        B = eotf_inverse_BT2020(B, is_12_bits_system=is_12_bits_system)

    Cbc = np.where((B - Yc) <= 0, (B - Yc) / 1.9404, (B - Yc) / 1.5816)
    Crc = np.where((R - Yc) <= 0, (R - Yc) / 1.7184, (R - Yc) / 0.9936)
    Yc *= Y_max - Y_min
    Yc += Y_min
    Cbc *= C_max - C_min
    Crc *= C_max - C_min
    Cbc += (C_max + C_min) / 2
    Crc += (C_max + C_min) / 2

    YcCbcCrc = tstack([Yc, Cbc, Crc])

    if out_int:
        return as_int_array(np.round(YcCbcCrc))
    else:
        return from_range_1(YcCbcCrc)


def YcCbcCrc_to_RGB(
    YcCbcCrc: ArrayLike,
    in_bits: Integer = 10,
    in_legal: Boolean = True,
    in_int: Boolean = False,
    is_12_bits_system: Boolean = False,
    **kwargs: Any,
) -> NDArray:
    """
    Convert an array of *Yc'Cbc'Crc'* colour encoding values to the
    corresponding *RGB* array of linear values.

    Parameters
    ----------
    YcCbcCrc
        Input *Yc'Cbc'Crc'* colour encoding array of linear float values.
    in_bits
        Bit depth for integer input, or used in the calculation of the
        denominator for legal range float values, i.e. 8-bit means the float
        value for legal white is *235 / 255*. Default is *10*.
    in_legal
        Whether to treat the input values as legal range. Default is *False*.
    in_int
        Whether to treat the input values as ``in_bits`` integer code values.
        Default is *False*.
    is_12_bits_system
        *Recommendation ITU-R BT.2020* EOTF (EOCF) adopts different parameters
        for 10 and 12 bit systems. Default is *False*.

    Other Parameters
    ----------------
    in_range
        Array overriding the computed range such as
        *in_range = (Y_min, Y_max, C_min, C_max)*. If ``in_range`` is
        undefined, *Y_min*, *Y_max*, *C_min* and *C_max* will be computed using
        :func:`colour.models.rgb.ycbcr.ranges_YCbCr` definition.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* array of linear float values.

    Notes
    -----
    +----------------+-----------------------+---------------+
    | **Domain \\***  | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``YcCbcCrc``   | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    +----------------+-----------------------+---------------+
    | **Range \\***   | **Scale - Reference** | **Scale - 1** |
    +================+=======================+===============+
    | ``RGB``        | [0, 1]                | [0, 1]        |
    +----------------+-----------------------+---------------+

    \\* This definition has input and output integer switches, thus the
    domain-range scale information is only given for the floating point mode.

    Warnings
    --------
    This definition is specifically for usage with
    *Recommendation ITU-R BT.2020* when adopting the constant luminance
    implementation.

    References
    ----------
    :cite:`InternationalTelecommunicationUnion2015h`,
    :cite:`Wikipedia2004d`

    Examples
    --------
    >>> YcCbcCrc = np.array([1689, 2048, 2048])
    >>> YcCbcCrc_to_RGB(YcCbcCrc, in_legal=True, in_bits=12, in_int=True,
    ...                 is_12_bits_system=True)
    ... # doctest: +ELLIPSIS
    array([ 0.1800903...,  0.1800903...,  0.1800903...])
    """

    if in_int:
        YcCbcCrc = as_float_array(YcCbcCrc)
    else:
        YcCbcCrc = to_domain_1(YcCbcCrc)

    Yc, Cbc, Crc = tsplit(YcCbcCrc)
    Y_min, Y_max, C_min, C_max = kwargs.get(
        "in_range", ranges_YCbCr(in_bits, in_legal, in_int)
    )

    Yc -= Y_min
    Cbc -= (C_max + C_min) / 2
    Crc -= (C_max + C_min) / 2
    Yc *= 1 / (Y_max - Y_min)
    Cbc *= 1 / (C_max - C_min)
    Crc *= 1 / (C_max - C_min)
    B = np.where(Cbc <= 0, Cbc * 1.9404 + Yc, Cbc * 1.5816 + Yc)
    R = np.where(Crc <= 0, Crc * 1.7184 + Yc, Crc * 0.9936 + Yc)

    with domain_range_scale("ignore"):
        Yc = as_float_array(eotf_BT2020(Yc, is_12_bits_system))
        B = as_float_array(eotf_BT2020(B, is_12_bits_system))
        R = as_float_array(eotf_BT2020(R, is_12_bits_system))

    G = (Yc - 0.0593 * B - 0.2627 * R) / 0.6780

    RGB = tstack([R, G, B])

    return from_range_1(RGB)



"""
Von Kries Chromatic Adaptation Model
====================================

Defines the *Von Kries* chromatic adaptation model objects:

-   :func:`colour.adaptation.matrix_chromatic_adaptation_VonKries`
-   :func:`colour.adaptation.chromatic_adaptation_VonKries`

References
----------
-   :cite:`Fairchild2013t` : Fairchild, M. D. (2013). Chromatic Adaptation
    Models. In Color Appearance Models (3rd ed., pp. 4179-4252). Wiley.
    ISBN:B00DAYO8E2
"""

from __future__ import annotations

import numpy as np

from colour.adaptation import CHROMATIC_ADAPTATION_TRANSFORMS
from colour.algebra import matrix_dot, vector_dot
from colour.hints import ArrayLike, Literal, NDArray, Union
from colour.utilities import (
    from_range_1,
    row_as_diagonal,
    to_domain_1,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "matrix_chromatic_adaptation_VonKries",
    "chromatic_adaptation_VonKries",
]


def matrix_chromatic_adaptation_VonKries(
    XYZ_w: ArrayLike,
    XYZ_wr: ArrayLike,
    transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
) -> NDArray:
    """
    Compute the *chromatic adaptation* matrix from test viewing conditions
    to reference viewing conditions.

    Parameters
    ----------
    XYZ_w
        Test viewing conditions *CIE XYZ* tristimulus values of whitepoint.
    XYZ_wr
        Reference viewing conditions *CIE XYZ* tristimulus values of
        whitepoint.
    transform
        Chromatic adaptation transform.

    Returns
    -------
    :class:`numpy.ndarray`
        Chromatic adaptation matrix :math:`M_{cat}`.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ_w``  | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_wr`` | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Fairchild2013t`

    Examples
    --------
    >>> XYZ_w = np.array([0.95045593, 1.00000000, 1.08905775])
    >>> XYZ_wr = np.array([0.96429568, 1.00000000, 0.82510460])
    >>> matrix_chromatic_adaptation_VonKries(XYZ_w, XYZ_wr)
    ... # doctest: +ELLIPSIS
    array([[ 1.0425738...,  0.0308910..., -0.0528125...],
           [ 0.0221934...,  1.0018566..., -0.0210737...],
           [-0.0011648..., -0.0034205...,  0.7617890...]])

    Using Bradford method:

    >>> XYZ_w = np.array([0.95045593, 1.00000000, 1.08905775])
    >>> XYZ_wr = np.array([0.96429568, 1.00000000, 0.82510460])
    >>> method = 'Bradford'
    >>> matrix_chromatic_adaptation_VonKries(XYZ_w, XYZ_wr, method)
    ... # doctest: +ELLIPSIS
    array([[ 1.0479297...,  0.0229468..., -0.0501922...],
           [ 0.0296278...,  0.9904344..., -0.0170738...],
           [-0.0092430...,  0.0150551...,  0.7518742...]])
    """

    XYZ_w = to_domain_1(XYZ_w)
    XYZ_wr = to_domain_1(XYZ_wr)

    transform = validate_method(
        transform,
        CHROMATIC_ADAPTATION_TRANSFORMS,
        '"{0}" chromatic adaptation transform is invalid, '
        "it must be one of {1}!",
    )

    M = CHROMATIC_ADAPTATION_TRANSFORMS[transform]

    RGB_w = np.einsum("...i,...ij->...j", XYZ_w, np.transpose(M))
    RGB_wr = np.einsum("...i,...ij->...j", XYZ_wr, np.transpose(M))

    D = RGB_wr / RGB_w

    D = row_as_diagonal(D)

    M_CAT = matrix_dot(np.linalg.inv(M), D)
    M_CAT = matrix_dot(M_CAT, M)

    return M_CAT


def chromatic_adaptation_VonKries(
    XYZ: ArrayLike,
    XYZ_w: ArrayLike,
    XYZ_wr: ArrayLike,
    transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
) -> NDArray:
    """
    Adapt given stimulus from test viewing conditions to reference viewing
    conditions.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values of stimulus to adapt.
    XYZ_w
        Test viewing conditions *CIE XYZ* tristimulus values of whitepoint.
    XYZ_wr
        Reference viewing conditions *CIE XYZ* tristimulus values of
        whitepoint.
    transform
        Chromatic adaptation transform.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE XYZ_c* tristimulus values of the stimulus corresponding colour.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_n``  | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+
    | ``XYZ_r``  | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ_c``  | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Fairchild2013t`

    Examples
    --------
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_w = np.array([0.95045593, 1.00000000, 1.08905775])
    >>> XYZ_wr = np.array([0.96429568, 1.00000000, 0.82510460])
    >>> chromatic_adaptation_VonKries(XYZ, XYZ_w, XYZ_wr)  # doctest: +ELLIPSIS
    array([ 0.2163881...,  0.1257    ,  0.0384749...])

    Using Bradford method:

    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> XYZ_w = np.array([0.95045593, 1.00000000, 1.08905775])
    >>> XYZ_wr = np.array([0.96429568, 1.00000000, 0.82510460])
    >>> transform = 'Bradford'
    >>> chromatic_adaptation_VonKries(XYZ, XYZ_w, XYZ_wr, transform)
    ... # doctest: +ELLIPSIS
    array([ 0.2166600...,  0.1260477...,  0.0385506...])
    """

    XYZ = to_domain_1(XYZ)

    M_CAT = matrix_chromatic_adaptation_VonKries(XYZ_w, XYZ_wr, transform)
    XYZ_a = vector_dot(M_CAT, XYZ)

    return from_range_1(XYZ_a)



"""Showcases *Nayatani (1995)* colour appearance model computations."""

import numpy as np

import colour
from colour.appearance.nayatani95 import CAM_ReferenceSpecification_Nayatani95
from colour.utilities import message_box

message_box('"Nayatani (1995)" Colour Appearance Model Computations')

XYZ = np.array([19.01, 20.00, 21.78])
XYZ_n = np.array([95.05, 100.00, 108.88])
Y_o = 20.0
E_o = 5000.0
E_or = 1000.0
message_box(
    f'Converting to the "Nayatani (1995)" colour appearance model '
    f"specification using given parameters:\n\n"
    f"\tXYZ: {XYZ}\n"
    f"\tXYZ_n: {XYZ_n}\n"
    f"\tY_o: {Y_o}\n"
    f"\tE_o: {E_o}\n"
    f"\tE_or: {E_or}"
)
specification = colour.XYZ_to_Nayatani95(XYZ, XYZ_n, Y_o, E_o, E_or)
print(specification)

print("\n")

message_box(
    'Broadcasting the current output "Nayatani (1995)" colour appearance '
    "model specification to the reference specification.\n"
    "The intent of this reference specification is to provide names "
    'as closest as possible to the "Mark D. Fairchild" reference.\n'
    "The current output specification is meant to be consistent with "
    "the other colour appearance model specification by using same "
    "argument names for consistency wherever possible."
)

print(CAM_ReferenceSpecification_Nayatani95(*specification.values))



"""
Interpolation
=============

Defines the classes and definitions for interpolating variables.

-   :class:`colour.KernelInterpolator`: 1-D function generic interpolation with
    arbitrary kernel.
-   :class:`colour.NearestNeighbourInterpolator`: 1-D function
    nearest-neighbour interpolation.
-   :class:`colour.LinearInterpolator`: 1-D function linear interpolation.
-   :class:`colour.SpragueInterpolator`: 1-D function fifth-order polynomial
    interpolation using *Sprague (1880)* method.
-   :class:`colour.CubicSplineInterpolator`: 1-D function cubic spline
    interpolation.
-   :class:`colour.PchipInterpolator`: 1-D function piecewise cube Hermite
    interpolation.
-   :class:`colour.NullInterpolator`: 1-D function null interpolation.
-   :func:`colour.lagrange_coefficients`: Computation of
    *Lagrange Coefficients*.
-   :func:`colour.algebra.table_interpolation_trilinear`: Trilinear
    interpolation with table.
-   :func:`colour.algebra.table_interpolation_tetrahedral`: Tetrahedral
    interpolation with table.
-   :attr:`colour.TABLE_INTERPOLATION_METHODS`: Supported table interpolation
    methods.
-   :func:`colour.table_interpolation`: Interpolation with table using given
    method.

References
----------
-   :cite:`Bourkeb` : Bourke, P. (n.d.). Trilinear Interpolation. Retrieved
    January 13, 2018, from http://paulbourke.net/miscellaneous/interpolation/
-   :cite:`Burger2009b` : Burger, W., & Burge, M. J. (2009). Principles of
    Digital Image Processing. Springer London. doi:10.1007/978-1-84800-195-4
-   :cite:`CIETC1-382005f` : CIE TC 1-38. (2005). 9.2.4 Method of
    interpolation for uniformly spaced independent variable. In CIE 167:2005
    Recommended Practice for Tabulating Spectral Data for Use in Colour
    Computations (pp. 1-27). ISBN:978-3-901906-41-1
-   :cite:`CIETC1-382005h` : CIE TC 1-38. (2005). Table V. Values of the
    c-coefficients of Equ.s 6 and 7. In CIE 167:2005 Recommended Practice for
    Tabulating Spectral Data for Use in Colour Computations (p. 19).
    ISBN:978-3-901906-41-1
-   :cite:`Fairman1985b` : Fairman, H. S. (1985). The calculation of weight
    factors for tristimulus integration. Color Research & Application, 10(4),
    199-203. doi:10.1002/col.5080100407
-   :cite:`Kirk2006` : Kirk, R. (2006). Truelight Software Library 2.0.
    Retrieved July 8, 2017, from
    https://www.filmlight.ltd.uk/pdf/whitepapers/FL-TL-TN-0057-SoftwareLib.pdf
-   :cite:`Westland2012h` : Westland, S., Ripamonti, C., & Cheung, V. (2012).
    Interpolation Methods. In Computational Colour Science Using MATLAB (2nd
    ed., pp. 29-37). ISBN:978-0-470-66569-5
-   :cite:`Wikipedia2003a` : Wikipedia. (2003). Lagrange polynomial -
    Definition. Retrieved January 20, 2016, from
    https://en.wikipedia.org/wiki/Lagrange_polynomial#Definition
-   :cite:`Wikipedia2005b` : Wikipedia. (2005). Lanczos resampling. Retrieved
    October 14, 2017, from https://en.wikipedia.org/wiki/Lanczos_resampling
"""

from __future__ import annotations

import itertools
import numpy as np
import scipy.interpolate
from collections.abc import Mapping
from functools import reduce

from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.hints import (
    Any,
    ArrayLike,
    Callable,
    Dict,
    DTypeNumber,
    Floating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Integer,
    Literal,
    NDArray,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
)
from colour.utilities import (
    CaseInsensitiveMapping,
    as_array,
    as_float_array,
    as_float,
    as_int_array,
    attest,
    closest_indexes,
    interval,
    is_numeric,
    optional,
    runtime_warning,
    tsplit,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "kernel_nearest_neighbour",
    "kernel_linear",
    "kernel_sinc",
    "kernel_lanczos",
    "kernel_cardinal_spline",
    "KernelInterpolator",
    "NearestNeighbourInterpolator",
    "LinearInterpolator",
    "SpragueInterpolator",
    "CubicSplineInterpolator",
    "PchipInterpolator",
    "NullInterpolator",
    "lagrange_coefficients",
    "vertices_and_relative_coordinates",
    "table_interpolation_trilinear",
    "table_interpolation_tetrahedral",
    "TABLE_INTERPOLATION_METHODS",
    "table_interpolation",
]


def kernel_nearest_neighbour(x: ArrayLike) -> NDArray:
    """
    Return the *nearest-neighbour* kernel evaluated at given samples.

    Parameters
    ----------
    x
        Samples at which to evaluate the *nearest-neighbour* kernel.

    Returns
    -------
    :class:`numpy.ndarray`
        The *nearest-neighbour* kernel evaluated at given samples.

    References
    ----------
    :cite:`Burger2009b`

    Examples
    --------
    >>> kernel_nearest_neighbour(np.linspace(0, 1, 10))
    array([1, 1, 1, 1, 1, 0, 0, 0, 0, 0])
    """

    return np.where(np.abs(x) < 0.5, 1, 0)


def kernel_linear(x: ArrayLike) -> NDArray:
    """
    Return the *linear* kernel evaluated at given samples.

    Parameters
    ----------
    x
        Samples at which to evaluate the *linear* kernel.

    Returns
    -------
    :class:`numpy.ndarray`
        The *linear* kernel evaluated at given samples.

    References
    ----------
    :cite:`Burger2009b`

    Examples
    --------
    >>> kernel_linear(np.linspace(0, 1, 10))  # doctest: +ELLIPSIS
    array([ 1.        ,  0.8888888...,  0.7777777...,  \
0.6666666...,  0.5555555...,
            0.4444444...,  0.3333333...,  0.2222222...,  \
0.1111111...,  0.        ])
    """

    return np.where(np.abs(x) < 1, 1 - np.abs(x), 0)


def kernel_sinc(x: ArrayLike, a: Floating = 3) -> NDArray:
    """
    Return the *sinc* kernel evaluated at given samples.

    Parameters
    ----------
    x
        Samples at which to evaluate the *sinc* kernel.
    a
        Size of the *sinc* kernel.

    Returns
    -------
    :class:`numpy.ndarray`
        The *sinc* kernel evaluated at given samples.

    References
    ----------
    :cite:`Burger2009b`

    Examples
    --------
    >>> kernel_sinc(np.linspace(0, 1, 10))  # doctest: +ELLIPSIS
    array([  1.0000000...e+00,   9.7981553...e-01,   9.2072542...e-01,
             8.2699334...e-01,   7.0531659...e-01,   5.6425327...e-01,
             4.1349667...e-01,   2.6306440...e-01,   1.2247694...e-01,
             3.8981718...e-17])
    """

    x = as_float_array(x)

    attest(bool(a >= 1), '"a" must be equal or superior to 1!')

    return np.where(np.abs(x) < a, np.sinc(x), 0)


def kernel_lanczos(x: ArrayLike, a: Floating = 3) -> NDArray:
    """
    Return the *lanczos* kernel evaluated at given samples.

    Parameters
    ----------
    x
        Samples at which to evaluate the *lanczos* kernel.
    a
        Size of the *lanczos* kernel.

    Returns
    -------
    :class:`numpy.ndarray`
        The *lanczos* kernel evaluated at given samples.

    References
    ----------
    :cite:`Wikipedia2005b`

    Examples
    --------
    >>> kernel_lanczos(np.linspace(0, 1, 10))  # doctest: +ELLIPSIS
    array([  1.0000000...e+00,   9.7760615...e-01,   9.1243770...e-01,
             8.1030092...e-01,   6.8012706...e-01,   5.3295773...e-01,
             3.8071690...e-01,   2.3492839...e-01,   1.0554054...e-01,
             3.2237621...e-17])
    """

    x = as_float_array(x)

    attest(bool(a >= 1), '"a" must be equal or superior to 1!')

    return np.where(np.abs(x) < a, np.sinc(x) * np.sinc(x / a), 0)


def kernel_cardinal_spline(
    x: ArrayLike, a: Floating = 0.5, b: Floating = 0.0
) -> NDArray:
    """
    Return the *cardinal spline* kernel evaluated at given samples.

    Notable *cardinal spline* :math:`a` and :math:`b` parameterizations:

    -   *Catmull-Rom*: :math:`(a=0.5, b=0)`
    -   *Cubic B-Spline*: :math:`(a=0, b=1)`
    -   *Mitchell-Netravalli*: :math:`(a=\\cfrac{1}{3}, b=\\cfrac{1}{3})`

    Parameters
    ----------
    x
        Samples at which to evaluate the *cardinal spline* kernel.
    a
        :math:`a` control parameter.
    b
        :math:`b` control parameter.

    Returns
    -------
    :class:`numpy.ndarray`
        The *cardinal spline* kernel evaluated at given samples.

    References
    ----------
    :cite:`Burger2009b`

    Examples
    --------
    >>> kernel_cardinal_spline(np.linspace(0, 1, 10))  # doctest: +ELLIPSIS
    array([ 1.        ,  0.9711934...,  0.8930041...,  \
0.7777777...,  0.6378600...,
            0.4855967...,  0.3333333...,  0.1934156...,  \
0.0781893...,  0.        ])
    """

    x = as_float_array(x)

    x_abs = np.abs(x)
    y = np.where(
        x_abs < 1,
        (-6 * a - 9 * b + 12) * x_abs**3
        + (6 * a + 12 * b - 18) * x_abs**2
        - 2 * b
        + 6,
        (-6 * a - b) * x_abs**3
        + (30 * a + 6 * b) * x_abs**2
        + (-48 * a - 12 * b) * x_abs
        + 24 * a
        + 8 * b,
    )
    y[x_abs >= 2] = 0

    return 1 / 6 * y


class KernelInterpolator:
    """
    Kernel based interpolation of a 1-D function.

    The reconstruction of a continuous signal can be described as a linear
    convolution operation. Interpolation can be expressed as a convolution of
    the given discrete function :math:`g(x)` with some continuous interpolation
    kernel :math:`k(w)`::

        :math:`\\hat{g}(w_0) = [k * g](w_0) = \
\\sum_{x=-\\infty}^{\\infty}k(w_0 - x)\\cdot g(x)`

    Parameters
    ----------
    x
        Independent :math:`x` variable values corresponding with :math:`y`
        variable.
    y
        Dependent and already known :math:`y` variable values to
        interpolate.
    window
        Width of the window in samples on each side.
    kernel
        Kernel to use for interpolation.
    kernel_kwargs
         Arguments to use when calling the kernel.
    padding_kwargs
         Arguments to use when padding :math:`y` variable values with the
         :func:`np.pad` definition.
    dtype
        Data type used for internal conversions.

    Attributes
    ----------
    -   :attr:`~colour.KernelInterpolator.x`
    -   :attr:`~colour.KernelInterpolator.y`
    -   :attr:`~colour.KernelInterpolator.window`
    -   :attr:`~colour.KernelInterpolator.kernel`
    -   :attr:`~colour.KernelInterpolator.kernel_kwargs`
    -   :attr:`~colour.KernelInterpolator.padding_kwargs`

    Methods
    -------
    -   :meth:`~colour.KernelInterpolator.__init__`
    -   :meth:`~colour.KernelInterpolator.__call__`

    References
    ----------
    :cite:`Burger2009b`, :cite:`Wikipedia2005b`

    Examples
    --------
    Interpolating a single numeric variable:

    >>> y = np.array([5.9200, 9.3700, 10.8135, 4.5100,
    ...               69.5900, 27.8007, 86.0500])
    >>> x = np.arange(len(y))
    >>> f = KernelInterpolator(x, y)
    >>> f(0.5)  # doctest: +ELLIPSIS
    6.9411400...

    Interpolating an `ArrayLike` variable:

    >>> f([0.25, 0.75])  # doctest: +ELLIPSIS
    array([ 6.1806208...,  8.0823848...])

    Using a different *lanczos* kernel:

    >>> f = KernelInterpolator(x, y, kernel=kernel_sinc)
    >>> f([0.25, 0.75])  # doctest: +ELLIPSIS
    array([ 6.5147317...,  8.3965466...])

    Using a different window size:

    >>> f = KernelInterpolator(
    ...     x,
    ...     y,
    ...     window=16,
    ...     kernel=kernel_lanczos,
    ...     kernel_kwargs={'a': 16})
    >>> f([0.25, 0.75])  # doctest: +ELLIPSIS
    array([ 5.3961792...,  5.6521093...])
    """

    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        window: Floating = 3,
        kernel: Callable = kernel_lanczos,
        kernel_kwargs: Optional[Dict] = None,
        padding_kwargs: Optional[Dict] = None,
        dtype: Optional[Type[DTypeNumber]] = None,
    ):
        dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

        self._x_p: NDArray = np.array([])
        self._y_p: NDArray = np.array([])

        self._x: NDArray = np.array([])
        self._y: NDArray = np.array([])
        self._window: Floating = 3
        self._padding_kwargs: Dict = {
            "pad_width": (window, window),
            "mode": "reflect",
        }
        self._kernel: Callable = kernel_lanczos
        self._kernel_kwargs: Dict = {}
        self._dtype: Type[DTypeNumber] = dtype

        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self.x = x  # type: ignore[assignment]
        self.y = y  # type: ignore[assignment]
        self.window = window
        self.padding_kwargs = optional(padding_kwargs, self._padding_kwargs)
        self.kernel = kernel
        self.kernel_kwargs = optional(kernel_kwargs, self._kernel_kwargs)

        self._validate_dimensions()

    @property
    def x(self) -> NDArray:
        """
        Getter and setter property for the independent :math:`x` variable.

        Parameters
        ----------
        value
            Value to set the independent :math:`x` variable with.

        Returns
        -------
        :class:`numpy.ndarray`
            Independent :math:`x` variable.
        """

        return self._x

    @x.setter
    def x(self, value: ArrayLike):
        """Setter for the **self.x** property."""

        value = np.atleast_1d(value).astype(self._dtype)

        attest(
            value.ndim == 1,
            '"x" independent variable must have exactly one dimension!',
        )

        value_interval = interval(value)

        if value_interval.size != 1:
            runtime_warning(
                '"x" independent variable is not uniform, '
                "unpredictable results may occur!"
            )

        self._x = as_array(value, self._dtype)

        self._x_p = np.pad(
            self._x,
            as_int_array([self._window, self._window]),
            "linear_ramp",
            end_values=(
                np.min(self._x) - self._window * value_interval[0],
                np.max(self._x) + self._window * value_interval[0],
            ),
        )

    @property
    def y(self) -> NDArray:
        """
        Getter and setter property for the dependent and already known
        :math:`y` variable.

        Parameters
        ----------
        value
            Value to set the dependent and already known :math:`y` variable
            with.

        Returns
        -------
        :class:`numpy.ndarray`
            Dependent and already known :math:`y` variable.
        """

        return self._y

    @y.setter
    def y(self, value: ArrayLike):
        """Setter for the **self.y** property."""

        value = np.atleast_1d(value).astype(self._dtype)

        attest(
            value.ndim == 1,
            '"y" dependent variable must have exactly one dimension!',
        )

        self._y = as_array(value, self._dtype)

        if self._window is not None:
            self._y_p = np.pad(self._y, **self._padding_kwargs)

    @property
    def window(self) -> Floating:
        """
        Getter and setter property for the window.

        Parameters
        ----------
        value
            Value to set the window with.

        Returns
        -------
        :class:`numpy.floating`
            Window.
        """

        return self._window

    @window.setter
    def window(self, value: Floating):
        """Setter for the **self.window** property."""

        attest(
            bool(value >= 1), '"window" must be equal to or greater than 1!'
        )

        self._window = value

        # Triggering "self._x_p" update.
        if self._x is not None:
            self.x = self._x

        # Triggering "self._y_p" update.
        if self._y is not None:
            self.y = self._y

    @property
    def kernel(self) -> Callable:
        """
        Getter and setter property for the kernel callable.

        Parameters
        ----------
        value
            Value to set the kernel callable.

        Returns
        -------
        Callable
            Kernel callable.
        """

        return self._kernel

    @kernel.setter
    def kernel(self, value: Callable):
        """Setter for the **self.kernel** property."""

        attest(
            hasattr(value, "__call__"),
            f'"kernel" property: "{value}" is not callable!',
        )

        self._kernel = value

    @property
    def kernel_kwargs(self) -> Dict:
        """
        Getter and setter property for the kernel call time arguments.

        Parameters
        ----------
        value
            Value to call the interpolation kernel with.

        Returns
        -------
        :class:`dict`
            Kernel call time arguments.
        """

        return self._kernel_kwargs

    @kernel_kwargs.setter
    def kernel_kwargs(self, value: Dict):
        """Setter for the **self.kernel_kwargs** property."""

        attest(
            isinstance(value, dict),
            f'"kernel_kwargs" property: "{value}" type is not "dict"!',
        )

        self._kernel_kwargs = value

    @property
    def padding_kwargs(self) -> Dict:
        """
        Getter and setter property for the kernel call time arguments.

        Parameters
        ----------
        value
            Value to call the interpolation kernel with.

        Returns
        -------
        :class:`dict`
            Kernel call time arguments.
        """

        return self._padding_kwargs

    @padding_kwargs.setter
    def padding_kwargs(self, value: Dict):
        """Setter for the **self.padding_kwargs** property."""

        attest(
            isinstance(value, Mapping),
            f'"padding_kwargs" property: "{value}" type is not a "Mapping" instance!',
        )

        self._padding_kwargs = value

        # Triggering "self._y_p" update.
        if self._y is not None:
            self.y = self._y

    def __call__(self, x: FloatingOrArrayLike) -> FloatingOrNDArray:
        """
        Evaluate the interpolator at given point(s).

        Parameters
        ----------
        x
            Point(s) to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Interpolated value(s).
        """

        x = np.atleast_1d(x).astype(self._dtype)

        xi = as_float(self._evaluate(x))

        return xi

    def _evaluate(self, x: NDArray) -> NDArray:
        """
        Perform the interpolator evaluation at given points.

        Parameters
        ----------
        x
            Points to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated points values.
        """

        self._validate_dimensions()
        self._validate_interpolation_range(x)

        x_interval = interval(self._x)[0]
        x_f = np.floor(x / x_interval)

        windows = x_f[:, np.newaxis] + np.arange(
            -self._window + 1, self._window + 1
        )
        clip_l = min(self._x_p) / x_interval
        clip_h = max(self._x_p) / x_interval
        windows = np.clip(windows, clip_l, clip_h) - clip_l
        windows = as_int_array(np.around(windows))

        return np.sum(
            self._y_p[windows]
            * self._kernel(
                x[:, np.newaxis] / x_interval
                - windows
                - min(self._x_p) / x_interval,
                **self._kernel_kwargs,
            ),
            axis=-1,
        )

    def _validate_dimensions(self):
        """Validate that the variables dimensions are the same."""

        if len(self._x) != len(self._y):
            raise ValueError(
                '"x" independent and "y" dependent variables have different '
                f'dimensions: "{len(self._x)}", "{len(self._y)}"'
            )

    def _validate_interpolation_range(self, x: NDArray):
        """Validate given point to be in interpolation range."""

        below_interpolation_range = x < self._x[0]
        above_interpolation_range = x > self._x[-1]

        if below_interpolation_range.any():
            raise ValueError(f'"{x}" is below interpolation range.')

        if above_interpolation_range.any():
            raise ValueError(f'"{x}" is above interpolation range.')


class NearestNeighbourInterpolator(KernelInterpolator):
    """
    A nearest-neighbour interpolator.

    Other Parameters
    ----------------
    dtype
        Data type used for internal conversions.
    padding_kwargs
         Arguments to use when padding :math:`y` variable values with the
         :func:`np.pad` definition.
    window
        Width of the window in samples on each side.
    x
        Independent :math:`x` variable values corresponding with :math:`y`
        variable.
    y
        Dependent and already known :math:`y` variable values to
        interpolate.

    Methods
    -------
    -   :meth:`~colour.NearestNeighbourInterpolator.__init__`
    """

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs["kernel"] = kernel_nearest_neighbour
        if "kernel_kwargs" in kwargs:
            del kwargs["kernel_kwargs"]

        super().__init__(*args, **kwargs)


class LinearInterpolator:
    """
    Interpolate linearly a 1-D function.

    Parameters
    ----------
    x
        Independent :math:`x` variable values corresponding with :math:`y`
        variable.
    y
        Dependent and already known :math:`y` variable values to
        interpolate.
    dtype
        Data type used for internal conversions.

    Attributes
    ----------
    -   :attr:`~colour.LinearInterpolator.x`
    -   :attr:`~colour.LinearInterpolator.y`

    Methods
    -------
    -   :meth:`~colour.LinearInterpolator.__init__`
    -   :meth:`~colour.LinearInterpolator.__call__`

    Notes
    -----
    -   This class is a wrapper around *numpy.interp* definition.

    Examples
    --------
    Interpolating a single numeric variable:

    >>> y = np.array([5.9200, 9.3700, 10.8135, 4.5100,
    ...               69.5900, 27.8007, 86.0500])
    >>> x = np.arange(len(y))
    >>> f = LinearInterpolator(x, y)
    >>> f(0.5)  # doctest: +ELLIPSIS
    7.64...

    Interpolating an `ArrayLike` variable:

    >>> f([0.25, 0.75])
    array([ 6.7825,  8.5075])
    """

    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        dtype: Optional[Type[DTypeNumber]] = None,
    ):
        dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

        self._x: NDArray = np.array([])
        self._y: NDArray = np.array([])
        self._dtype: Type[DTypeNumber] = dtype

        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self.x = x  # type: ignore[assignment]
        self.y = y  # type: ignore[assignment]

        self._validate_dimensions()

    @property
    def x(self) -> NDArray:
        """
        Getter and setter property for the independent :math:`x` variable.

        Parameters
        ----------
        value
            Value to set the independent :math:`x` variable with.

        Returns
        -------
        :class:`numpy.ndarray`
            Independent :math:`x` variable.
        """

        return self._x

    @x.setter
    def x(self, value: ArrayLike):
        """Setter for the **self.x** property."""

        value = np.atleast_1d(value).astype(self._dtype)

        attest(
            value.ndim == 1,
            '"x" independent variable must have exactly one dimension!',
        )

        self._x = value

    @property
    def y(self) -> NDArray:
        """
        Getter and setter property for the dependent and already known
        :math:`y` variable.

        Parameters
        ----------
        value
            Value to set the dependent and already known :math:`y` variable
            with.

        Returns
        -------
        :class:`numpy.ndarray`
            Dependent and already known :math:`y` variable.
        """

        return self._y

    @y.setter
    def y(self, value: ArrayLike):
        """Setter for the **self.y** property."""

        value = np.atleast_1d(value).astype(self._dtype)

        attest(
            value.ndim == 1,
            '"y" dependent variable must have exactly one dimension!',
        )

        self._y = value

    def __call__(self, x: FloatingOrArrayLike) -> FloatingOrNDArray:
        """
        Evaluate the interpolating polynomial at given point(s).


        Parameters
        ----------
        x
            Point(s) to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Interpolated value(s).
        """

        x = np.atleast_1d(x).astype(self._dtype)

        xi = as_float(self._evaluate(x))

        return xi

    def _evaluate(self, x: NDArray) -> NDArray:
        """
        Perform the interpolating polynomial evaluation at given points.

        Parameters
        ----------
        x
            Points to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated points values.
        """

        self._validate_dimensions()
        self._validate_interpolation_range(x)

        return np.interp(x, self._x, self._y)

    def _validate_dimensions(self):
        """Validate that the variables dimensions are the same."""

        if len(self._x) != len(self._y):
            raise ValueError(
                '"x" independent and "y" dependent variables have different '
                f'dimensions: "{len(self._x)}", "{len(self._y)}"'
            )

    def _validate_interpolation_range(self, x: NDArray):
        """Validate given point to be in interpolation range."""

        below_interpolation_range = x < self._x[0]
        above_interpolation_range = x > self._x[-1]

        if below_interpolation_range.any():
            raise ValueError(f'"{x}" is below interpolation range.')

        if above_interpolation_range.any():
            raise ValueError(f'"{x}" is above interpolation range.')


class SpragueInterpolator:
    """
    Construct a fifth-order polynomial that passes through :math:`y` dependent
    variable.

    *Sprague (1880)* method is recommended by the *CIE* for interpolating
    functions having a uniformly spaced independent variable.

    Parameters
    ----------
    x
        Independent :math:`x` variable values corresponding with :math:`y`
        variable.
    y
        Dependent and already known :math:`y` variable values to
        interpolate.
    dtype
        Data type used for internal conversions.

    Attributes
    ----------
    -   :attr:`~colour.SpragueInterpolator.x`
    -   :attr:`~colour.SpragueInterpolator.y`

    Methods
    -------
    -   :meth:`~colour.SpragueInterpolator.__init__`
    -   :meth:`~colour.SpragueInterpolator.__call__`

    Notes
    -----
    -   The minimum number :math:`k` of data points required along the
        interpolation axis is :math:`k=6`.

    References
    ----------
    :cite:`CIETC1-382005f`, :cite:`Westland2012h`

    Examples
    --------
    Interpolating a single numeric variable:

    >>> y = np.array([5.9200, 9.3700, 10.8135, 4.5100,
    ...               69.5900, 27.8007, 86.0500])
    >>> x = np.arange(len(y))
    >>> f = SpragueInterpolator(x, y)
    >>> f(0.5)  # doctest: +ELLIPSIS
    7.2185025...

    Interpolating an `ArrayLike` variable:

    >>> f([0.25, 0.75])  # doctest: +ELLIPSIS
    array([ 6.7295161...,  7.8140625...])
    """

    SPRAGUE_C_COEFFICIENTS = np.array(
        [
            [884, -1960, 3033, -2648, 1080, -180],
            [508, -540, 488, -367, 144, -24],
            [-24, 144, -367, 488, -540, 508],
            [-180, 1080, -2648, 3033, -1960, 884],
        ]
    )
    """
    Defines the coefficients used to generate extra points for boundaries
    interpolation.

    SPRAGUE_C_COEFFICIENTS, (4, 6)

    References
    ----------
    :cite:`CIETC1-382005h`
    """

    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        dtype: Optional[Type[DTypeNumber]] = None,
    ):
        dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

        self._xp: NDArray = np.array([])
        self._yp: NDArray = np.array([])

        self._x: NDArray = np.array([])
        self._y: NDArray = np.array([])
        self._dtype: Type[DTypeNumber] = dtype

        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self.x = x  # type: ignore[assignment]
        self.y = y  # type: ignore[assignment]

        self._validate_dimensions()

    @property
    def x(self) -> NDArray:
        """
        Getter and setter property for the independent :math:`x` variable.

        Parameters
        ----------
        value
            Value to set the independent :math:`x` variable with.

        Returns
        -------
        :class:`numpy.ndarray`
            Independent :math:`x` variable.
        """

        return self._x

    @x.setter
    def x(self, value: ArrayLike):
        """Setter for the **self.x** property."""

        value = as_array(np.atleast_1d(value), self._dtype)

        attest(
            value.ndim == 1,
            '"x" independent variable must have exactly one dimension!',
        )

        self._x = value

        value_interval = interval(self._x)[0]

        xp1 = self._x[0] - value_interval * 2
        xp2 = self._x[0] - value_interval
        xp3 = self._x[-1] + value_interval
        xp4 = self._x[-1] + value_interval * 2

        self._xp = np.concatenate(
            [
                as_array([xp1, xp2], self._dtype),
                value,
                as_array([xp3, xp4], self._dtype),
            ]
        )

    @property
    def y(self) -> NDArray:
        """
        Getter and setter property for the dependent and already known
        :math:`y` variable.

        Parameters
        ----------
        value
            Value to set the dependent and already known :math:`y` variable
            with.

        Returns
        -------
        :class:`numpy.ndarray`
            Dependent and already known :math:`y` variable.
        """

        return self._y

    @y.setter
    def y(self, value: ArrayLike):
        """Setter for the **self.y** property."""

        value = as_array(np.atleast_1d(value), self._dtype)

        attest(
            value.ndim == 1,
            '"y" dependent variable must have exactly one dimension!',
        )

        attest(
            len(value) >= 6,
            '"y" dependent variable values count must be equal to or '
            "greater than 6!",
        )

        self._y = value

        yp1 = np.ravel(
            (
                np.dot(
                    self.SPRAGUE_C_COEFFICIENTS[0],
                    np.array(value[0:6]).reshape([6, 1]),
                )
            )
            / 209
        )[0]
        yp2 = np.ravel(
            (
                np.dot(
                    self.SPRAGUE_C_COEFFICIENTS[1],
                    np.array(value[0:6]).reshape([6, 1]),
                )
            )
            / 209
        )[0]
        yp3 = np.ravel(
            (
                np.dot(
                    self.SPRAGUE_C_COEFFICIENTS[2],
                    np.array(value[-6:]).reshape([6, 1]),
                )
            )
            / 209
        )[0]
        yp4 = np.ravel(
            (
                np.dot(
                    self.SPRAGUE_C_COEFFICIENTS[3],
                    np.array(value[-6:]).reshape([6, 1]),
                )
            )
            / 209
        )[0]

        self._yp = np.concatenate(
            [
                as_array([yp1, yp2], self._dtype),
                value,
                as_array([yp3, yp4], self._dtype),
            ]
        )

    def __call__(self, x: FloatingOrArrayLike) -> FloatingOrNDArray:
        """
        Evaluate the interpolating polynomial at given point(s).

        Parameters
        ----------
        x
            Point(s) to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Interpolated value(s).
        """

        x = np.atleast_1d(x).astype(self._dtype)

        xi = as_float(self._evaluate(x))

        return xi

    def _evaluate(self, x: NDArray) -> NDArray:
        """
        Perform the interpolating polynomial evaluation at given point.

        Parameters
        ----------
        x
            Point to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated point values.
        """

        self._validate_dimensions()
        self._validate_interpolation_range(x)

        i = np.searchsorted(self._xp, x) - 1
        X = (x - self._xp[i]) / (self._xp[i + 1] - self._xp[i])

        r = self._yp

        a0p = r[i]
        a1p = (
            2 * r[i - 2] - 16 * r[i - 1] + 16 * r[i + 1] - 2 * r[i + 2]
        ) / 24
        a2p = (
            -r[i - 2] + 16 * r[i - 1] - 30 * r[i] + 16 * r[i + 1] - r[i + 2]
        ) / 24
        a3p = (
            -9 * r[i - 2]
            + 39 * r[i - 1]
            - 70 * r[i]
            + 66 * r[i + 1]
            - 33 * r[i + 2]
            + 7 * r[i + 3]
        ) / 24
        a4p = (
            13 * r[i - 2]
            - 64 * r[i - 1]
            + 126 * r[i]
            - 124 * r[i + 1]
            + 61 * r[i + 2]
            - 12 * r[i + 3]
        ) / 24
        a5p = (
            -5 * r[i - 2]
            + 25 * r[i - 1]
            - 50 * r[i]
            + 50 * r[i + 1]
            - 25 * r[i + 2]
            + 5 * r[i + 3]
        ) / 24

        y = (
            a0p
            + a1p * X
            + a2p * X**2
            + a3p * X**3
            + a4p * X**4
            + a5p * X**5
        )

        return y

    def _validate_dimensions(self):
        """Validate that the variables dimensions are the same."""

        if len(self._x) != len(self._y):
            raise ValueError(
                '"x" independent and "y" dependent variables have different '
                f'dimensions: "{len(self._x)}", "{len(self._y)}"'
            )

    def _validate_interpolation_range(self, x: NDArray):
        """Validate given point to be in interpolation range."""

        below_interpolation_range = x < self._x[0]
        above_interpolation_range = x > self._x[-1]

        if below_interpolation_range.any():
            raise ValueError(f'"{x}" is below interpolation range.')

        if above_interpolation_range.any():
            raise ValueError(f'"{x}" is above interpolation range.')


class CubicSplineInterpolator(scipy.interpolate.interp1d):
    """
    Interpolate a 1-D function using cubic spline interpolation.

    Methods
    -------
    -   :meth:`~colour.CubicSplineInterpolator.__init__`

    Notes
    -----
    -   This class is a wrapper around *scipy.interpolate.interp1d* class.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(kind="cubic", *args, **kwargs)


class PchipInterpolator(scipy.interpolate.PchipInterpolator):
    """
    Interpolate a 1-D function using Piecewise Cubic Hermite Interpolating
    Polynomial interpolation.

    Attributes
    ----------
    -   :attr:`~colour.PchipInterpolator.y`

    Methods
    -------
    -   :meth:`~colour.PchipInterpolator.__init__`

    Notes
    -----
    -   This class is a wrapper around *scipy.interpolate.PchipInterpolator*
        class.
    """

    def __init__(self, x: ArrayLike, y: ArrayLike, *args: Any, **kwargs: Any):
        super().__init__(x, y, *args, **kwargs)

        self._y: NDArray = as_float_array(y)

    @property
    def y(self) -> NDArray:
        """
        Getter property for the dependent and already known :math:`y`
        variable.

        Returns
        -------
        :class:`numpy.ndarray`
            Dependent and already known :math:`y` variable.
        """

        return self._y


class NullInterpolator:
    """
    Perform 1-D function null interpolation, i.e. a call within given
    tolerances will return existing :math:`y` variable values and ``default``
    if outside tolerances.

    Parameters
    ----------
    x
        Independent :math:`x` variable values corresponding with :math:`y`
        variable.
    y
        Dependent and already known :math:`y` variable values to
        interpolate.
    absolute_tolerance
        Absolute tolerance.
    relative_tolerance
        Relative tolerance.
    default
        Default value for interpolation outside tolerances.
    dtype
        Data type used for internal conversions.

    Attributes
    ----------
    -   :attr:`~colour.NullInterpolator.x`
    -   :attr:`~colour.NullInterpolator.y`
    -   :attr:`~colour.NullInterpolator.relative_tolerance`
    -   :attr:`~colour.NullInterpolator.absolute_tolerance`
    -   :attr:`~colour.NullInterpolator.default`

    Methods
    -------
    -   :meth:`~colour.NullInterpolator.__init__`
    -   :meth:`~colour.NullInterpolator.__call__`

    Examples
    --------
    >>> y = np.array([5.9200, 9.3700, 10.8135, 4.5100,
    ...               69.5900, 27.8007, 86.0500])
    >>> x = np.arange(len(y))
    >>> f = NullInterpolator(x, y)
    >>> f(0.5)
    nan
    >>> f(1.0)  # doctest: +ELLIPSIS
    9.3699999...
    >>> f = NullInterpolator(x, y, absolute_tolerance=0.01)
    >>> f(1.01)  # doctest: +ELLIPSIS
    9.3699999...
    """

    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        absolute_tolerance: Floating = 10e-7,
        relative_tolerance: Floating = 10e-7,
        default: Floating = np.nan,
        dtype: Optional[Type[DTypeNumber]] = None,
    ):
        dtype = cast(Type[DTypeNumber], optional(dtype, DEFAULT_FLOAT_DTYPE))

        self._x: NDArray = np.array([])
        self._y: NDArray = np.array([])
        self._absolute_tolerance: float = 10e-7
        self._relative_tolerance: float = 10e-7
        self._default: Floating = np.nan
        self._dtype: Type[DTypeNumber] = dtype

        # TODO: Remove pragma when https://github.com/python/mypy/issues/3004
        # is resolved.
        self.x = x  # type: ignore[assignment]
        self.y = y  # type: ignore[assignment]
        self.absolute_tolerance = absolute_tolerance
        self.relative_tolerance = relative_tolerance
        self.default = default

        self._validate_dimensions()

    @property
    def x(self) -> NDArray:
        """
        Getter and setter property for the independent :math:`x` variable.

        Parameters
        ----------
        value
            Value to set the independent :math:`x` variable with.

        Returns
        -------
        :class:`numpy.ndarray`
            Independent :math:`x` variable.
        """

        return self._x

    @x.setter
    def x(self, value: ArrayLike):
        """Setter for the **self.x** property."""

        value = np.atleast_1d(value).astype(self._dtype)

        attest(
            value.ndim == 1,
            '"x" independent variable must have exactly one dimension!',
        )

        self._x = value

    @property
    def y(self) -> NDArray:
        """
        Getter and setter property for the dependent and already known
        :math:`y` variable.

        Parameters
        ----------
        value
            Value to set the dependent and already known :math:`y` variable
            with.

        Returns
        -------
        :class:`numpy.ndarray`
            Dependent and already known :math:`y` variable.
        """

        return self._y

    @y.setter
    def y(self, value: ArrayLike):
        """Setter for the **self.y** property."""

        value = np.atleast_1d(value).astype(self._dtype)

        attest(
            value.ndim == 1,
            '"y" dependent variable must have exactly one dimension!',
        )

        self._y = value

    @property
    def relative_tolerance(self) -> Floating:
        """
        Getter and setter property for the relative tolerance.

        Parameters
        ----------
        value
            Value to set the relative tolerance with.

        Returns
        -------
        :class:`numpy.floating`
            Relative tolerance.
        """

        return self._relative_tolerance

    @relative_tolerance.setter
    def relative_tolerance(self, value: Floating):
        """Setter for the **self.relative_tolerance** property."""

        attest(
            is_numeric(value),
            '"relative_tolerance" variable must be a "numeric"!',
        )

        self._relative_tolerance = float(value)

    @property
    def absolute_tolerance(self) -> Floating:
        """
        Getter and setter property for the absolute tolerance.

        Parameters
        ----------
        value
            Value to set the absolute tolerance with.

        Returns
        -------
        :class:`numpy.floating`
            Absolute tolerance.
        """

        return self._absolute_tolerance

    @absolute_tolerance.setter
    def absolute_tolerance(self, value: Floating):
        """Setter for the **self.absolute_tolerance** property."""

        attest(
            is_numeric(value),
            '"absolute_tolerance" variable must be a "numeric"!',
        )

        self._absolute_tolerance = float(value)

    @property
    def default(self) -> Floating:
        """
        Getter and setter property for the default value for call outside
        tolerances.

        Parameters
        ----------
        value
            Value to set the default value with.

        Returns
        -------
        :class:`numpy.floating`
            Default value.
        """

        return self._default

    @default.setter
    def default(self, value: Floating):
        """Setter for the **self.default** property."""

        attest(is_numeric(value), '"default" variable must be a "numeric"!')

        self._default = value

    def __call__(self, x: FloatingOrArrayLike) -> FloatingOrNDArray:
        """
        Evaluate the interpolator at given point(s).


        Parameters
        ----------
        x
            Point(s) to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Interpolated value(s).
        """

        x = np.atleast_1d(x).astype(self._dtype)

        xi = as_float(self._evaluate(x))

        return xi

    def _evaluate(self, x: NDArray) -> NDArray:
        """
        Perform the interpolator evaluation at given points.

        Parameters
        ----------
        x
            Points to evaluate the interpolant at.

        Returns
        -------
        :class:`numpy.ndarray`
            Interpolated points values.
        """

        self._validate_dimensions()
        self._validate_interpolation_range(x)

        indexes = closest_indexes(self._x, x)
        values = self._y[indexes]
        values[
            ~np.isclose(
                self._x[indexes],
                x,
                rtol=self._absolute_tolerance,
                atol=self._relative_tolerance,
            )
        ] = self._default

        return values

    def _validate_dimensions(self):
        """Validate that the variables dimensions are the same."""

        if len(self._x) != len(self._y):
            raise ValueError(
                '"x" independent and "y" dependent variables have different '
                f'dimensions: "{len(self._x)}", "{len(self._y)}"'
            )

    def _validate_interpolation_range(self, x: NDArray):
        """Validate given point to be in interpolation range."""

        below_interpolation_range = x < self._x[0]
        above_interpolation_range = x > self._x[-1]

        if below_interpolation_range.any():
            raise ValueError(f'"{x}" is below interpolation range.')

        if above_interpolation_range.any():
            raise ValueError(f'"{x}" is above interpolation range.')


def lagrange_coefficients(r: Floating, n: Integer = 4) -> NDArray:
    """
    Compute the *Lagrange Coefficients* at given point :math:`r` for degree
    :math:`n`.

    Parameters
    ----------
    r
        Point to get the *Lagrange Coefficients* at.
    n
        Degree of the *Lagrange Coefficients* being calculated.

    Returns
    -------
    :class:`numpy.ndarray`

    References
    ----------
    :cite:`Fairman1985b`, :cite:`Wikipedia2003a`

    Examples
    --------
    >>> lagrange_coefficients(0.1)
    array([ 0.8265,  0.2755, -0.1305,  0.0285])
    """

    r_i = np.arange(n)
    L_n = []
    for j in range(len(r_i)):
        basis = [
            (r - r_i[i]) / (r_i[j] - r_i[i]) for i in range(len(r_i)) if i != j
        ]
        L_n.append(reduce(lambda x, y: x * y, basis))  # noqa

    return np.array(L_n)


def vertices_and_relative_coordinates(
    V_xyz: ArrayLike, table: ArrayLike
) -> Tuple[NDArray, NDArray]:
    """
    Compute the vertices coordinates and indexes relative :math:`V_{xyzr}`
    coordinates from given :math:`V_{xyzr}` values and interpolation table.

    Parameters
    ----------
    V_xyz
        :math:`V_{xyz}` values to transform to indexes relative
        :math:`V_{xyzr}` values.
    table
        4-Dimensional (NxNxNx3) interpolation table.

    Returns
    -------
    :class:`tuple`
        Vertices coordinates and indexes relative :math:`V_{xyzr}` coordinates.

    Examples
    --------
    >>> import os
    >>> import colour
    >>> path = os.path.join(
    ...     os.path.dirname(__file__),'..', 'io', 'luts', 'tests', 'resources',
    ...     'iridas_cube', 'Colour_Correct.cube')
    >>> LUT = colour.read_LUT(path)
    >>> table = LUT.table
    >>> prng = np.random.RandomState(4)
    >>> V_xyz = colour.algebra.random_triplet_generator(3, random_state=prng)
    >>> print(V_xyz)  # doctest: +ELLIPSIS
    [[ 0.9670298...  0.7148159...  0.9762744...]
     [ 0.5472322...  0.6977288...  0.0062302...]
     [ 0.9726843...  0.2160895...  0.2529823...]]
    >>> vertices, V_xyzr = vertices_and_relative_coordinates(V_xyz, table)
    >>> print(vertices)
    [[[ 0.833311  0.833311  0.833311]
      [ 0.349416  0.657749  0.041083]
      [ 0.797894 -0.035412 -0.035412]]
    <BLANKLINE>
     [[ 0.833311  0.833311  1.249963]
      [ 0.340435  0.743769  0.340435]
      [ 0.752767 -0.028479  0.362144]]
    <BLANKLINE>
     [[ 0.707102  1.110435  0.707102]
      [ 0.344991  1.050213 -0.007621]
      [ 0.633333  0.316667  0.      ]]
    <BLANKLINE>
     [[ 0.519714  0.744729  0.744729]
      [ 0.314204  1.120871  0.314204]
      [ 0.732278  0.315626  0.315626]]
    <BLANKLINE>
     [[ 1.06561   0.648957  0.648957]
      [ 0.589195  0.589195  0.139164]
      [ 1.196841 -0.053117 -0.053117]]
    <BLANKLINE>
     [[ 1.        0.666667  1.      ]
      [ 0.594601  0.594601  0.369586]
      [ 1.162588 -0.050372  0.353948]]
    <BLANKLINE>
     [[ 0.894606  0.894606  0.66959 ]
      [ 0.663432  0.930188  0.12992 ]
      [ 1.038439  0.310899 -0.05287 ]]
    <BLANKLINE>
     [[ 1.249966  1.249966  1.249966]
      [ 0.682749  0.991082  0.374416]
      [ 1.131225  0.29792   0.29792 ]]]
    >>> print(V_xyzr)  # doctest: +ELLIPSIS
    [[ 0.9010895...  0.1444479...  0.9288233...]
     [ 0.6416967...  0.0931864...  0.0186907...]
     [ 0.9180530...  0.6482684...  0.7589470...]]
    """

    V_xyz = np.clip(V_xyz, 0, 1)
    table = as_float_array(table)

    V_xyz = np.reshape(V_xyz, (-1, 3))

    # Indexes computations where ``i_m`` is the maximum index value on a given
    # table axis, ``i_f`` and ``i_c`` respectively the floor and ceiling
    # indexes encompassing a given V_xyz value.
    i_m = np.array(table.shape[0:-1]) - 1
    i_f = as_int_array(np.floor(V_xyz * i_m))
    i_f = np.clip(i_f, 0, i_m)
    i_c = np.clip(i_f + 1, 0, i_m)

    # Relative to indexes ``V_xyz`` values.
    V_xyzr = i_m * V_xyz - i_f

    i_f_c = i_f, i_c

    # Vertices computations by indexing ``table`` with the ``i_f`` and ``i_c``
    # indexes. 8 encompassing vertices are computed for a given V_xyz value
    # forming a cube around it:
    vertices = np.array(
        [
            table[
                i_f_c[i[0]][..., 0], i_f_c[i[1]][..., 1], i_f_c[i[2]][..., 2]
            ]
            for i in itertools.product(*zip([0, 0, 0], [1, 1, 1]))
        ]
    )

    return vertices, V_xyzr


def table_interpolation_trilinear(
    V_xyz: ArrayLike, table: ArrayLike
) -> NDArray:
    """
    Perform the trilinear interpolation of given :math:`V_{xyz}` values using
    given interpolation table.

    Parameters
    ----------
    V_xyz
        :math:`V_{xyz}` values to interpolate.
    table
        4-Dimensional (NxNxNx3) interpolation table.

    Returns
    -------
    :class:`numpy.ndarray`
        Interpolated :math:`V_{xyz}` values.

    References
    ----------
    :cite:`Bourkeb`

    Examples
    --------
    >>> import os
    >>> import colour
    >>> path = os.path.join(
    ...     os.path.dirname(__file__),'..', 'io', 'luts', 'tests', 'resources',
    ...     'iridas_cube', 'Colour_Correct.cube')
    >>> LUT = colour.read_LUT(path)
    >>> table = LUT.table
    >>> prng = np.random.RandomState(4)
    >>> V_xyz = colour.algebra.random_triplet_generator(3, random_state=prng)
    >>> print(V_xyz)  # doctest: +ELLIPSIS
    [[ 0.9670298...  0.7148159...  0.9762744...]
     [ 0.5472322...  0.6977288...  0.0062302...]
     [ 0.9726843...  0.2160895...  0.2529823...]]
    >>> table_interpolation_trilinear(V_xyz, table)  # doctest: +ELLIPSIS
    array([[ 1.0120664...,  0.7539146...,  1.0228540...],
           [ 0.5075794...,  0.6479459...,  0.1066404...],
           [ 1.0976519...,  0.1785998...,  0.2299897...]])
    """

    V_xyz = as_float_array(V_xyz)

    vertices, V_xyzr = vertices_and_relative_coordinates(V_xyz, table)

    vertices = np.moveaxis(vertices, 0, 1)
    x, y, z = (f[:, np.newaxis] for f in tsplit(V_xyzr))

    weights = np.moveaxis(
        np.transpose(
            [
                (1 - x) * (1 - y) * (1 - z),
                (1 - x) * (1 - y) * z,
                (1 - x) * y * (1 - z),
                (1 - x) * y * z,
                x * (1 - y) * (1 - z),
                x * (1 - y) * z,
                x * y * (1 - z),
                x * y * z,
            ]
        ),
        0,
        -1,
    )

    xyz_o = np.reshape(np.sum(vertices * weights, 1), V_xyz.shape)

    return xyz_o


def table_interpolation_tetrahedral(
    V_xyz: ArrayLike, table: ArrayLike
) -> NDArray:
    """
    Perform the tetrahedral interpolation of given :math:`V_{xyz}` values using
    given interpolation table.

    Parameters
    ----------
    V_xyz
        :math:`V_{xyz}` values to interpolate.
    table
        4-Dimensional (NxNxNx3) interpolation table.

    Returns
    -------
    :class:`numpy.ndarray`
        Interpolated :math:`V_{xyz}` values.

    References
    ----------
    :cite:`Kirk2006`

    Examples
    --------
    >>> import os
    >>> import colour
    >>> path = os.path.join(
    ...     os.path.dirname(__file__),'..', 'io', 'luts', 'tests', 'resources',
    ...     'iridas_cube', 'Colour_Correct.cube')
    >>> LUT = colour.read_LUT(path)
    >>> table = LUT.table
    >>> prng = np.random.RandomState(4)
    >>> V_xyz = colour.algebra.random_triplet_generator(3, random_state=prng)
    >>> print(V_xyz)  # doctest: +ELLIPSIS
    [[ 0.9670298...  0.7148159...  0.9762744...]
     [ 0.5472322...  0.6977288...  0.0062302...]
     [ 0.9726843...  0.2160895...  0.2529823...]]
    >>> table_interpolation_tetrahedral(V_xyz, table)  # doctest: +ELLIPSIS
    array([[ 1.0196197...,  0.7674062...,  1.0311751...],
           [ 0.5105603...,  0.6466722...,  0.1077296...],
           [ 1.1178206...,  0.1762039...,  0.2209534...]])
    """

    V_xyz = as_float_array(V_xyz)

    vertices, V_xyzr = vertices_and_relative_coordinates(V_xyz, table)

    vertices = np.moveaxis(vertices, 0, -1)
    V000, V001, V010, V011, V100, V101, V110, V111 = tsplit(vertices)
    x, y, z = (r[:, np.newaxis] for r in tsplit(V_xyzr))

    xyz_o = np.select(
        [
            np.logical_and(x > y, y > z),
            np.logical_and(x > y, x > z),
            np.logical_and(x > y, np.logical_and(y <= z, x <= z)),
            np.logical_and(x <= y, z > y),
            np.logical_and(x <= y, z > x),
            np.logical_and(x <= y, np.logical_and(z <= y, z <= x)),
        ],
        [
            (1 - x) * V000 + (x - y) * V100 + (y - z) * V110 + z * V111,
            (1 - x) * V000 + (x - z) * V100 + (z - y) * V101 + y * V111,
            (1 - z) * V000 + (z - x) * V001 + (x - y) * V101 + y * V111,
            (1 - z) * V000 + (z - y) * V001 + (y - x) * V011 + x * V111,
            (1 - y) * V000 + (y - z) * V010 + (z - x) * V011 + x * V111,
            (1 - y) * V000 + (y - x) * V010 + (x - z) * V110 + z * V111,
        ],
    )

    xyz_o = np.reshape(xyz_o, V_xyz.shape)

    return xyz_o


TABLE_INTERPOLATION_METHODS = CaseInsensitiveMapping(
    {
        "Trilinear": table_interpolation_trilinear,
        "Tetrahedral": table_interpolation_tetrahedral,
    }
)
TABLE_INTERPOLATION_METHODS.__doc__ = """
Supported table interpolation methods.

References
----------
:cite:`Bourkeb`, :cite:`Kirk2006`
"""


def table_interpolation(
    V_xyz: ArrayLike,
    table: ArrayLike,
    method: Union[Literal["Trilinear", "Tetrahedral"], str] = "Trilinear",
) -> NDArray:
    """
    Perform interpolation of given :math:`V_{xyz}` values using given
    interpolation table.

    Parameters
    ----------
    V_xyz
        :math:`V_{xyz}` values to interpolate.
    table
        4-Dimensional (NxNxNx3) interpolation table.
    method
        Interpolation method.

    Returns
    -------
    :class:`numpy.ndarray`
        Interpolated :math:`V_{xyz}` values.

    References
    ----------
    :cite:`Bourkeb`, :cite:`Kirk2006`

    Examples
    --------
    >>> import os
    >>> import colour
    >>> path = os.path.join(
    ...     os.path.dirname(__file__),'..', 'io', 'luts', 'tests', 'resources',
    ...     'iridas_cube', 'Colour_Correct.cube')
    >>> LUT = colour.read_LUT(path)
    >>> table = LUT.table
    >>> prng = np.random.RandomState(4)
    >>> V_xyz = colour.algebra.random_triplet_generator(3, random_state=prng)
    >>> print(V_xyz)  # doctest: +ELLIPSIS
    [[ 0.9670298...  0.7148159...  0.9762744...]
     [ 0.5472322...  0.6977288...  0.0062302...]
     [ 0.9726843...  0.2160895...  0.2529823...]]
    >>> table_interpolation(V_xyz, table)  # doctest: +ELLIPSIS
    array([[ 1.0120664...,  0.7539146...,  1.0228540...],
           [ 0.5075794...,  0.6479459...,  0.1066404...],
           [ 1.0976519...,  0.1785998...,  0.2299897...]])
    >>> table_interpolation(V_xyz, table, method='Tetrahedral')
    ... # doctest: +ELLIPSIS
    array([[ 1.0196197...,  0.7674062...,  1.0311751...],
           [ 0.5105603...,  0.6466722...,  0.1077296...],
           [ 1.1178206...,  0.1762039...,  0.2209534...]])
    """

    method = validate_method(method, TABLE_INTERPOLATION_METHODS)

    return TABLE_INTERPOLATION_METHODS[method](V_xyz, table)



"""
Robertson (1968) Correlated Colour Temperature
==============================================

Defines the *Robertson (1968)* correlated colour temperature :math:`T_{cp}`
computations objects:

-   :func:`colour.temperature.uv_to_CCT_Robertson1968`: Correlated colour
    temperature :math:`T_{cp}` and :math:`\\Delta_{uv}` computation of given
    *CIE UCS* colourspace *uv* chromaticity coordinates using *Robertson
    (1968)* method.
-   :func:`colour.temperature.CCT_to_uv_Robertson1968`: *CIE UCS* colourspace
    *uv* chromaticity coordinates computation of given correlated colour
    temperature :math:`T_{cp}` and :math:`\\Delta_{uv}` using *Robertson
    (1968)* method.

References
----------
-   :cite:`AdobeSystems2013` : Adobe Systems. (2013). Adobe DNG Software
    Development Kit (SDK) - 1.3.0.0 -
    dng_sdk_1_3/dng_sdk/source/dng_temperature.cpp::dng_temperature::\
Set_xy_coord. https://www.adobe.com/support/downloads/dng/dng_sdk.html
-   :cite:`AdobeSystems2013a` : Adobe Systems. (2013). Adobe DNG Software
    Development Kit (SDK) - 1.3.0.0 -
    dng_sdk_1_3/dng_sdk/source/dng_temperature.cpp::dng_temperature::xy_coord.
    https://www.adobe.com/support/downloads/dng/dng_sdk.html
-   :cite:`Wyszecki2000x` : Wyszecki, Günther, & Stiles, W. S. (2000). Table
    1(3.11) Isotemperature Lines. In Color Science: Concepts and Methods,
    Quantitative Data and Formulae (p. 228). Wiley. ISBN:978-0-471-39918-6
-   :cite:`Wyszecki2000y` : Wyszecki, Günther, & Stiles, W. S. (2000).
    DISTRIBUTION TEMPERATURE, COLOR TEMPERATURE, AND CORRELATED COLOR
    TEMPERATURE. In Color Science: Concepts and Methods, Quantitative Data and
    Formulae (pp. 224-229). Wiley. ISBN:978-0-471-39918-6
"""

from __future__ import annotations

import numpy as np
from dataclasses import dataclass

from colour.hints import ArrayLike, Floating, List, NDArray, Tuple
from colour.utilities import as_float_array, tsplit

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "DATA_ISOTEMPERATURE_LINES_ROBERTSON1968",
    "ISOTemperatureLine_Specification_Robertson1968",
    "ISOTEMPERATURE_LINES_ROBERTSON1968",
    "uv_to_CCT_Robertson1968",
    "CCT_to_uv_Robertson1968",
]

DATA_ISOTEMPERATURE_LINES_ROBERTSON1968: Tuple = (
    (0, 0.18006, 0.26352, -0.24341),
    (10, 0.18066, 0.26589, -0.25479),
    (20, 0.18133, 0.26846, -0.26876),
    (30, 0.18208, 0.27119, -0.28539),
    (40, 0.18293, 0.27407, -0.30470),
    (50, 0.18388, 0.27709, -0.32675),
    (60, 0.18494, 0.28021, -0.35156),
    (70, 0.18611, 0.28342, -0.37915),
    (80, 0.18740, 0.28668, -0.40955),
    (90, 0.18880, 0.28997, -0.44278),
    (100, 0.19032, 0.29326, -0.47888),
    (125, 0.19462, 0.30141, -0.58204),
    (150, 0.19962, 0.30921, -0.70471),
    (175, 0.20525, 0.31647, -0.84901),
    (200, 0.21142, 0.32312, -1.0182),
    (225, 0.21807, 0.32909, -1.2168),
    (250, 0.22511, 0.33439, -1.4512),
    (275, 0.23247, 0.33904, -1.7298),
    (300, 0.24010, 0.34308, -2.0637),
    (325, 0.24792, 0.34655, -2.4681),  # 0.24702 --> 0.24792 Bruce Lindbloom
    (350, 0.25591, 0.34951, -2.9641),
    (375, 0.26400, 0.35200, -3.5814),
    (400, 0.27218, 0.35407, -4.3633),
    (425, 0.28039, 0.35577, -5.3762),
    (450, 0.28863, 0.35714, -6.7262),
    (475, 0.29685, 0.35823, -8.5955),
    (500, 0.30505, 0.35907, -11.324),
    (525, 0.31320, 0.35968, -15.628),
    (550, 0.32129, 0.36011, -23.325),
    (575, 0.32931, 0.36038, -40.770),
    (600, 0.33724, 0.36051, -116.45),
)
"""
*Robertson (1968)* iso-temperature lines as a *tuple* as follows::

    (
        ('Reciprocal Megakelvin', 'CIE 1960 Chromaticity Coordinate *u*',
         'CIE 1960 Chromaticity Coordinate *v*', 'Slope'),
        ...,
        ('Reciprocal Megakelvin', 'CIE 1960 Chromaticity Coordinate *u*',
         'CIE 1960 Chromaticity Coordinate *v*', 'Slope'),
    )

Notes
-----
-   A correction has been done by Lindbloom for *325* Megakelvin
    temperature: 0.24702 --> 0.24792

References
----------
:cite:`Wyszecki2000x`
"""


@dataclass
class ISOTemperatureLine_Specification_Robertson1968:
    """
    Define the data for a *Roberston (1968)* iso-temperature line.

    Parameters
    ----------
    r
        Temperature :math:`r` in reciprocal mega-kelvin degrees.
    u
        *u* chromaticity coordinate of the temperature :math:`r`.
    v
        *v* chromaticity coordinate of the temperature :math:`r`.
    t
        Slope of the *v* chromaticity coordinate.
    """

    r: Floating
    u: Floating
    v: Floating
    t: Floating


ISOTEMPERATURE_LINES_ROBERTSON1968: List = [
    ISOTemperatureLine_Specification_Robertson1968(*x)
    for x in DATA_ISOTEMPERATURE_LINES_ROBERTSON1968
]


def _uv_to_CCT_Robertson1968(uv: ArrayLike) -> NDArray:
    """
    Return the correlated colour temperature :math:`T_{cp}` and
    :math:`\\Delta_{uv}` from given *CIE UCS* colourspace *uv* chromaticity
    coordinates using *Roberston (1968)* method.

    Parameters
    ----------
    uv
        *CIE UCS* colourspace *uv* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        Correlated colour temperature :math:`T_{cp}`, :math:`\\Delta_{uv}`.
    """

    u, v = tsplit(uv)

    last_dt = last_dv = last_du = 0

    for i in range(1, 31):
        wr_ruvt = ISOTEMPERATURE_LINES_ROBERTSON1968[i]
        wr_ruvt_previous = ISOTEMPERATURE_LINES_ROBERTSON1968[i - 1]

        du = 1
        dv = wr_ruvt.t

        length = np.hypot(1, dv)

        du /= length
        dv /= length

        uu = u - wr_ruvt.u
        vv = v - wr_ruvt.v

        dt = -uu * dv + vv * du

        if dt <= 0 or i == 30:
            if dt > 0:
                dt = 0

            dt = -dt

            f = 0 if i == 1 else dt / (last_dt + dt)

            T = 1.0e6 / (wr_ruvt_previous.r * f + wr_ruvt.r * (1 - f))

            uu = u - (wr_ruvt_previous.u * f + wr_ruvt.u * (1 - f))
            vv = v - (wr_ruvt_previous.v * f + wr_ruvt.v * (1 - f))

            du = du * (1 - f) + last_du * f
            dv = dv * (1 - f) + last_dv * f

            length = np.hypot(du, dv)

            du /= length
            dv /= length

            D_uv = uu * du + vv * dv

            break

        last_dt = dt
        last_du = du
        last_dv = dv

    return np.array([T, -D_uv])


def uv_to_CCT_Robertson1968(uv: ArrayLike) -> NDArray:
    """
    Return the correlated colour temperature :math:`T_{cp}` and
    :math:`\\Delta_{uv}` from given *CIE UCS* colourspace *uv* chromaticity
    coordinates using *Roberston (1968)* method.

    Parameters
    ----------
    uv
        *CIE UCS* colourspace *uv* chromaticity coordinates.

    Returns
    -------
    :class:`numpy.ndarray`
        Correlated colour temperature :math:`T_{cp}`, :math:`\\Delta_{uv}`.

    References
    ----------
    :cite:`AdobeSystems2013`, :cite:`Wyszecki2000y`

    Examples
    --------
    >>> uv = np.array([0.193741375998230, 0.315221043940594])
    >>> uv_to_CCT_Robertson1968(uv)  # doctest: +ELLIPSIS
    array([  6.5000162...e+03,   8.3333289...e-03])
    """

    uv = as_float_array(uv)

    CCT_D_uv = [_uv_to_CCT_Robertson1968(a) for a in np.reshape(uv, (-1, 2))]

    return np.reshape(as_float_array(CCT_D_uv), uv.shape)


def _CCT_to_uv_Robertson1968(CCT_D_uv: ArrayLike) -> NDArray:
    """
    Return the *CIE UCS* colourspace *uv* chromaticity coordinates from given
    correlated colour temperature :math:`T_{cp}` and :math:`\\Delta_{uv}` using
    *Roberston (1968)* method.

    Parameters
    ----------
    CCT_D_uv
        Correlated colour temperature :math:`T_{cp}`, :math:`\\Delta_{uv}`.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE UCS* colourspace *uv* chromaticity coordinates.
    """

    CCT, D_uv = tsplit(CCT_D_uv)

    r = 1.0e6 / CCT

    u, v = np.nan, np.nan
    for i in range(30):
        wr_ruvt = ISOTEMPERATURE_LINES_ROBERTSON1968[i]
        wr_ruvt_next = ISOTEMPERATURE_LINES_ROBERTSON1968[i + 1]

        if r < wr_ruvt_next.r or i == 29:
            f = (wr_ruvt_next.r - r) / (wr_ruvt_next.r - wr_ruvt.r)

            u = wr_ruvt.u * f + wr_ruvt_next.u * (1 - f)
            v = wr_ruvt.v * f + wr_ruvt_next.v * (1 - f)

            uu1 = uu2 = 1.0
            vv1, vv2 = wr_ruvt.t, wr_ruvt_next.t

            length1 = np.hypot(1, vv1)
            length2 = np.hypot(1, vv2)

            uu1 /= length1
            vv1 /= length1

            uu2 /= length2
            vv2 /= length2

            uu3 = uu1 * f + uu2 * (1 - f)
            vv3 = vv1 * f + vv2 * (1 - f)

            len3 = np.sqrt(uu3 * uu3 + vv3 * vv3)

            uu3 /= len3
            vv3 /= len3

            u += uu3 * -D_uv
            v += vv3 * -D_uv

            break

    return np.array([u, v])


def CCT_to_uv_Robertson1968(CCT_D_uv: ArrayLike) -> NDArray:
    """
    Return the *CIE UCS* colourspace *uv* chromaticity coordinates from given
    correlated colour temperature :math:`T_{cp}` and :math:`\\Delta_{uv}` using
    *Roberston (1968)* method.

    Parameters
    ----------
    CCT_D_uv
        Correlated colour temperature :math:`T_{cp}`, :math:`\\Delta_{uv}`.

    Returns
    -------
    :class:`numpy.ndarray`
        *CIE UCS* colourspace *uv* chromaticity coordinates.

    References
    ----------
    :cite:`AdobeSystems2013a`, :cite:`Wyszecki2000y`

    Examples
    --------
    >>> CCT_D_uv = np.array([6500.0081378199056, 0.008333331244225])
    >>> CCT_to_uv_Robertson1968(CCT_D_uv)  # doctest: +ELLIPSIS
    array([ 0.1937413...,  0.3152210...])
    """

    CCT_D_uv = as_float_array(CCT_D_uv)

    uv = [_CCT_to_uv_Robertson1968(a) for a in np.reshape(CCT_D_uv, (-1, 2))]

    return np.reshape(as_float_array(uv), CCT_D_uv.shape)



"""
Meng et al. (2015) - Reflectance Recovery
=========================================

Defines the objects for reflectance recovery using
*Meng, Simon and Hanika (2015)* method:

-   :func:`colour.recovery.XYZ_to_sd_Meng2015`

References
----------
-   :cite:`Meng2015c` : Meng, J., Simon, F., Hanika, J., & Dachsbacher, C.
    (2015). Physically Meaningful Rendering using Tristimulus Colours. Computer
    Graphics Forum, 34(4), 31-40. doi:10.1111/cgf.12676
"""

from __future__ import annotations

import numpy as np
from scipy.optimize import minimize

from colour.colorimetry import (
    MultiSpectralDistributions,
    SpectralDistribution,
    SpectralShape,
    handle_spectral_arguments,
    sd_ones,
    sd_to_XYZ_integration,
)
from colour.hints import ArrayLike, Dict, FloatingOrNDArray, NDArray, Optional
from colour.utilities import to_domain_1, from_range_100

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "SPECTRAL_SHAPE_MENG2015",
    "XYZ_to_sd_Meng2015",
]

SPECTRAL_SHAPE_MENG2015: SpectralShape = SpectralShape(360, 780, 5)
"""
Spectral shape according to *ASTM E308-15* practise shape but using an interval
of 5.
"""


def XYZ_to_sd_Meng2015(
    XYZ: ArrayLike,
    cmfs: Optional[MultiSpectralDistributions] = None,
    illuminant: Optional[SpectralDistribution] = None,
    optimisation_kwargs: Optional[Dict] = None,
) -> SpectralDistribution:
    """
    Recover the spectral distribution of given *CIE XYZ* tristimulus values
    using *Meng et al. (2015)* method.

    Parameters
    ----------
    XYZ
        *CIE XYZ* tristimulus values to recover the spectral distribution from.
    cmfs
        Standard observer colour matching functions. The wavelength
        :math:`\\lambda_{i}` range interval of the colour matching functions
        affects directly the time the computations take. The current default
        interval of 5 is a good compromise between precision and time spent,
        default to the *CIE 1931 2 Degree Standard Observer*.
    illuminant
        Illuminant spectral distribution, default to
        *CIE Standard Illuminant D65*.
    optimisation_kwargs
        Parameters for :func:`scipy.optimize.minimize` definition.

    Returns
    -------
    :class:`colour.SpectralDistribution`
        Recovered spectral distribution.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``XYZ``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    -   The definition used to convert spectrum to *CIE XYZ* tristimulus
        values is :func:`colour.colorimetry.spectral_to_XYZ_integration`
        definition because it processes any measurement interval opposed to
        :func:`colour.colorimetry.sd_to_XYZ_ASTME308` definition that
        handles only measurement interval of 1, 5, 10 or 20nm.

    References
    ----------
    :cite:`Meng2015c`

    Examples
    --------
    >>> from colour import MSDS_CMFS, SDS_ILLUMINANTS
    >>> from colour.utilities import numpy_print_options
    >>> XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
    >>> cmfs = (
    ...     MSDS_CMFS['CIE 1931 2 Degree Standard Observer'].
    ...     copy().align(SpectralShape(360, 780, 10))
    ... )
    >>> illuminant = SDS_ILLUMINANTS['D65'].copy().align(cmfs.shape)
    >>> sd = XYZ_to_sd_Meng2015(XYZ, cmfs, illuminant)
    >>> with numpy_print_options(suppress=True):
    ...     sd  # doctest: +SKIP
    SpectralDistribution([[ 360.        ,    0.0762005...],
                          [ 370.        ,    0.0761792...],
                          [ 380.        ,    0.0761363...],
                          [ 390.        ,    0.0761194...],
                          [ 400.        ,    0.0762539...],
                          [ 410.        ,    0.0761671...],
                          [ 420.        ,    0.0754649...],
                          [ 430.        ,    0.0731519...],
                          [ 440.        ,    0.0676701...],
                          [ 450.        ,    0.0577800...],
                          [ 460.        ,    0.0441993...],
                          [ 470.        ,    0.0285064...],
                          [ 480.        ,    0.0138728...],
                          [ 490.        ,    0.0033585...],
                          [ 500.        ,    0.       ...],
                          [ 510.        ,    0.       ...],
                          [ 520.        ,    0.       ...],
                          [ 530.        ,    0.       ...],
                          [ 540.        ,    0.0055767...],
                          [ 550.        ,    0.0317581...],
                          [ 560.        ,    0.0754491...],
                          [ 570.        ,    0.1314115...],
                          [ 580.        ,    0.1937649...],
                          [ 590.        ,    0.2559311...],
                          [ 600.        ,    0.3123173...],
                          [ 610.        ,    0.3584966...],
                          [ 620.        ,    0.3927335...],
                          [ 630.        ,    0.4159458...],
                          [ 640.        ,    0.4306660...],
                          [ 650.        ,    0.4391040...],
                          [ 660.        ,    0.4439497...],
                          [ 670.        ,    0.4463618...],
                          [ 680.        ,    0.4474625...],
                          [ 690.        ,    0.4479868...],
                          [ 700.        ,    0.4482116...],
                          [ 710.        ,    0.4482800...],
                          [ 720.        ,    0.4483472...],
                          [ 730.        ,    0.4484251...],
                          [ 740.        ,    0.4484633...],
                          [ 750.        ,    0.4485071...],
                          [ 760.        ,    0.4484969...],
                          [ 770.        ,    0.4484853...],
                          [ 780.        ,    0.4485134...]],
                         interpolator=SpragueInterpolator,
                         interpolator_kwargs={},
                         extrapolator=Extrapolator,
                         extrapolator_kwargs={...})
    >>> sd_to_XYZ_integration(sd, cmfs, illuminant) / 100  # doctest: +ELLIPSIS
    array([ 0.2065400...,  0.1219722...,  0.0513695...])
    """

    XYZ = to_domain_1(XYZ)

    cmfs, illuminant = handle_spectral_arguments(
        cmfs, illuminant, shape_default=SPECTRAL_SHAPE_MENG2015
    )

    sd = sd_ones(cmfs.shape)

    def objective_function(a: ArrayLike) -> FloatingOrNDArray:
        """Define the objective function."""

        return np.sum(np.diff(a) ** 2)

    def constraint_function(a: ArrayLike) -> NDArray:
        """Define the constraint function."""

        sd[:] = a
        return (
            sd_to_XYZ_integration(sd, cmfs=cmfs, illuminant=illuminant) - XYZ
        )

    wavelengths = sd.wavelengths
    bins = wavelengths.size

    optimisation_settings = {
        "method": "SLSQP",
        "constraints": {"type": "eq", "fun": constraint_function},
        "bounds": np.tile(np.array([0, 1000]), (bins, 1)),
        "options": {
            "ftol": 1e-10,
        },
    }
    if optimisation_kwargs is not None:
        optimisation_settings.update(optimisation_kwargs)

    result = minimize(objective_function, sd.values, **optimisation_settings)

    if not result.success:
        raise RuntimeError(
            f"Optimization failed for {XYZ} after {result.nit} iterations: "
            f'"{result.message}".'
        )

    return SpectralDistribution(
        from_range_100(result.x * 100),
        wavelengths,
        name=f"{XYZ} (XYZ) - Meng (2015)",
    )



"""Defines the unit tests for the :mod:`colour.models.osa_ucs` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.models import XYZ_to_OSA_UCS, OSA_UCS_to_XYZ
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_OSA_UCS",
    "TestOSA_UCS_to_XYZ",
]


class TestXYZ_to_OSA_UCS(unittest.TestCase):
    """
    Define :func:`colour.models.osa_ucs.XYZ_to_OSA_UCS` definition unit tests
    methods.
    """

    def test_XYZ_to_OSA_UCS(self):
        """Test :func:`colour.models.osa_ucs.XYZ_to_OSA_UCS` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_OSA_UCS(
                np.array([0.20654008, 0.12197225, 0.05136952]) * 100
            ),
            np.array([-3.00499790, 2.99713697, -9.66784231]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_OSA_UCS(
                np.array([0.14222010, 0.23042768, 0.10495772]) * 100
            ),
            np.array([-1.64657491, 4.59201565, 5.31738757]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_OSA_UCS(
                np.array([0.07818780, 0.06157201, 0.28099326]) * 100
            ),
            np.array([-5.08589672, -7.91062749, 0.98107575]),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_OSA_UCS(self):
        """
        Test :func:`colour.models.osa_ucs.XYZ_to_OSA_UCS` definition
        n-dimensional support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952]) * 100
        Ljg = XYZ_to_OSA_UCS(XYZ)

        XYZ = np.tile(XYZ, (6, 1))
        Ljg = np.tile(Ljg, (6, 1))
        np.testing.assert_almost_equal(XYZ_to_OSA_UCS(XYZ), Ljg, decimal=7)

        XYZ = np.reshape(XYZ, (2, 3, 3))
        Ljg = np.reshape(Ljg, (2, 3, 3))
        np.testing.assert_almost_equal(XYZ_to_OSA_UCS(XYZ), Ljg, decimal=7)

    def test_domain_range_scale_XYZ_to_OSA_UCS(self):
        """
        Test :func:`colour.models.osa_ucs.XYZ_to_OSA_UCS` definition domain
        and range scale support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952]) * 100
        Ljg = XYZ_to_OSA_UCS(XYZ)

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_OSA_UCS(XYZ * factor), Ljg * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_OSA_UCS(self):
        """
        Test :func:`colour.models.osa_ucs.XYZ_to_OSA_UCS` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ_to_OSA_UCS(np.array(case))


class TestOSA_UCS_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.osa_ucs.OSA_UCS_to_XYZ` definition unit tests
    methods.
    """

    def test_OSA_UCS_to_XYZ(self):
        """Test :func:`colour.models.osa_ucs.OSA_UCS_to_XYZ` definition."""

        np.testing.assert_allclose(
            OSA_UCS_to_XYZ(
                np.array([-3.00499790, 2.99713697, -9.66784231]),
                {"disp": False},
            ),
            np.array([0.20654008, 0.12197225, 0.05136952]) * 100,
            rtol=0.00001,
            atol=0.00001,
        )

        np.testing.assert_allclose(
            OSA_UCS_to_XYZ(
                np.array([-1.64657491, 4.59201565, 5.31738757]),
                {"disp": False},
            ),
            np.array([0.14222010, 0.23042768, 0.10495772]) * 100,
            rtol=0.00001,
            atol=0.00001,
        )

        np.testing.assert_allclose(
            OSA_UCS_to_XYZ(
                np.array([-5.08589672, -7.91062749, 0.98107575]),
                {"disp": False},
            ),
            np.array([0.07818780, 0.06157201, 0.28099326]) * 100,
            rtol=0.00001,
            atol=0.00001,
        )

    def test_n_dimensional_OSA_UCS_to_XYZ(self):
        """
        Test :func:`colour.models.osa_ucs.OSA_UCS_to_XYZ` definition
        n-dimensional support.
        """

        Ljg = np.array([-3.00499790, 2.99713697, -9.66784231])
        XYZ = OSA_UCS_to_XYZ(Ljg)

        Ljg = np.tile(Ljg, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_allclose(
            OSA_UCS_to_XYZ(Ljg), XYZ, rtol=0.00001, atol=0.00001
        )

        Ljg = np.reshape(Ljg, (2, 3, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_allclose(
            OSA_UCS_to_XYZ(Ljg), XYZ, rtol=0.00001, atol=0.00001
        )

    def test_domain_range_scale_OSA_UCS_to_XYZ(self):
        """
        Test :func:`colour.models.osa_ucs.OSA_UCS_to_XYZ` definition domain
        and range scale support.
        """

        Ljg = np.array([-3.00499790, 2.99713697, -9.66784231])
        XYZ = OSA_UCS_to_XYZ(Ljg)

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    OSA_UCS_to_XYZ(Ljg * factor), XYZ * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_OSA_UCS_to_XYZ(self):
        """
        Test :func:`colour.models.osa_ucs.OSA_UCS_to_XYZ` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            OSA_UCS_to_XYZ(np.array(case))


if __name__ == "__main__":
    unittest.main()
